from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    Union,
)

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_graph import AssetGraph, InternalAssetGraph
from dagster._core.definitions.assets_job import (
    ASSET_BASE_JOB_PREFIX,
)
from dagster._core.definitions.cacheable_assets import AssetsDefinitionCacheableData
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.executor_definition import ExecutorDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.metadata import MetadataMapping
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import check_valid_name
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.instance import DagsterInstance
from dagster._serdes import whitelist_for_serdes
from dagster._utils import hash_collection

from .repository_data import CachingRepositoryData, RepositoryData
from .valid_definitions import (
    RepositoryListDefinition as RepositoryListDefinition,
)

if TYPE_CHECKING:
    from dagster._core.definitions import AssetsDefinition
    from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition
    from dagster._core.storage.asset_value_loader import AssetValueLoader


@whitelist_for_serdes
class RepositoryLoadData(
    NamedTuple(
        "_RepositoryLoadData",
        [
            ("cached_data_by_key", Mapping[str, Sequence[AssetsDefinitionCacheableData]]),
        ],
    )
):
    def __new__(cls, cached_data_by_key: Mapping[str, Sequence[AssetsDefinitionCacheableData]]):
        return super(RepositoryLoadData, cls).__new__(
            cls,
            cached_data_by_key=(
                check.mapping_param(
                    cached_data_by_key,
                    "cached_data_by_key",
                    key_type=str,
                    value_type=list,
                )
            ),
        )

    # Allow this to be hashed for use in `lru_cache`. This is needed because:
    # - `ReconstructableJob` uses `lru_cache`
    # - `ReconstructableJob` has a `ReconstructableRepository` attribute
    # - `ReconstructableRepository` has a `RepositoryLoadData` attribute
    # - `RepositoryLoadData` has collection attributes that are unhashable by default
    def __hash__(self) -> int:
        if not hasattr(self, "_hash"):
            self._hash = hash_collection(self)
        return self._hash


class RepositoryDefinition:
    """Define a repository that contains a group of definitions.

    Users should typically not create objects of this class directly. Instead, use the
    :py:func:`@repository` decorator.

    Args:
        name (str): The name of the repository.
        repository_data (RepositoryData): Contains the definitions making up the repository.
        description (Optional[str]): A string description of the repository.
        metadata (Optional[MetadataMapping]): A map of arbitrary metadata for the repository.
    """

    def __init__(
        self,
        name,
        *,
        repository_data,
        description=None,
        metadata=None,
        repository_load_data=None,
    ):
        self._name = check_valid_name(name)
        self._description = check.opt_str_param(description, "description")
        self._repository_data: RepositoryData = check.inst_param(
            repository_data, "repository_data", RepositoryData
        )
        self._metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
        self._repository_load_data = check.opt_inst_param(
            repository_load_data, "repository_load_data", RepositoryLoadData
        )

    @property
    def repository_load_data(self) -> Optional[RepositoryLoadData]:
        return self._repository_load_data

    @public
    @property
    def name(self) -> str:
        """str: The name of the repository."""
        return self._name

    @public
    @property
    def description(self) -> Optional[str]:
        """Optional[str]: A human-readable description of the repository."""
        return self._description

    @public
    @property
    def metadata(self) -> Optional[MetadataMapping]:
        """Optional[MetadataMapping]: Arbitrary metadata for the repository."""
        return self._metadata

    def load_all_definitions(self):
        # force load of all lazy constructed code artifacts
        self._repository_data.load_all_definitions()

    @public
    @property
    def job_names(self) -> Sequence[str]:
        """List[str]: Names of all jobs in the repository."""
        return self._repository_data.get_job_names()

    def get_top_level_resources(self) -> Mapping[str, ResourceDefinition]:
        return self._repository_data.get_top_level_resources()

    def get_env_vars_by_top_level_resource(self) -> Mapping[str, AbstractSet[str]]:
        return self._repository_data.get_env_vars_by_top_level_resource()

    def get_resource_key_mapping(self) -> Mapping[int, str]:
        return self._repository_data.get_resource_key_mapping()

    @public
    def has_job(self, name: str) -> bool:
        """Check if a job with a given name is present in the repository.

        Args:
            name (str): The name of the job.

        Returns:
            bool
        """
        return self._repository_data.has_job(name)

    @public
    def get_job(self, name: str) -> JobDefinition:
        """Get a job by name.

        If this job is present in the lazily evaluated dictionary passed to the
        constructor, but has not yet been constructed, only this job is constructed, and
        will be cached for future calls.

        Args:
            name (str): Name of the job to retrieve.

        Returns:
            JobDefinition: The job definition corresponding to
            the given name.
        """
        return self._repository_data.get_job(name)

    @public
    def get_all_jobs(self) -> Sequence[JobDefinition]:
        """Return all jobs in the repository as a list.

        Note that this will construct any job in the lazily evaluated dictionary that has
        not yet been constructed.

        Returns:
            List[JobDefinition]: All jobs in the repository.
        """
        return self._repository_data.get_all_jobs()

    @public
    @property
    def schedule_defs(self) -> Sequence[ScheduleDefinition]:
        """List[ScheduleDefinition]: All schedules in the repository."""
        return self._repository_data.get_all_schedules()

    @public
    def get_schedule_def(self, name: str) -> ScheduleDefinition:
        """Get a schedule definition by name.

        Args:
            name (str): The name of the schedule.

        Returns:
            ScheduleDefinition: The schedule definition.
        """
        return self._repository_data.get_schedule(name)

    @public
    def has_schedule_def(self, name: str) -> bool:
        """bool: Check if a schedule with a given name is present in the repository."""
        return self._repository_data.has_schedule(name)

    @public
    @property
    def sensor_defs(self) -> Sequence[SensorDefinition]:
        """Sequence[SensorDefinition]: All sensors in the repository."""
        return self._repository_data.get_all_sensors()

    @public
    def get_sensor_def(self, name: str) -> SensorDefinition:
        """Get a sensor definition by name.

        Args:
            name (str): The name of the sensor.

        Returns:
            SensorDefinition: The sensor definition.
        """
        return self._repository_data.get_sensor(name)

    @public
    def has_sensor_def(self, name: str) -> bool:
        """bool: Check if a sensor with a given name is present in the repository."""
        return self._repository_data.has_sensor(name)

    @property
    def source_assets_by_key(self) -> Mapping[AssetKey, SourceAsset]:
        return self._repository_data.get_source_assets_by_key()

    @property
    def assets_defs_by_key(self) -> Mapping[AssetKey, "AssetsDefinition"]:
        return self._repository_data.get_assets_defs_by_key()

    def has_implicit_global_asset_job_def(self) -> bool:
        """Returns true is there is a single implicit asset job for all asset keys in a repository."""
        return self.has_job(ASSET_BASE_JOB_PREFIX)

    def get_implicit_global_asset_job_def(self) -> JobDefinition:
        """A useful conveninence method for repositories where there are a set of assets with
        the same partitioning schema and one wants to access their corresponding implicit job
        easily.
        """
        if not self.has_job(ASSET_BASE_JOB_PREFIX):
            raise DagsterInvariantViolationError(
                "There is no single global asset job, likely due to assets using "
                "different partitioning schemes via their partitions_def parameter. You must "
                "use get_implicit_job_def_for_assets in order to access the correct implicit job."
            )

        return self.get_job(ASSET_BASE_JOB_PREFIX)

    def get_implicit_asset_job_names(self) -> Sequence[str]:
        return [
            job_name for job_name in self.job_names if job_name.startswith(ASSET_BASE_JOB_PREFIX)
        ]

    def get_implicit_job_def_for_assets(
        self, asset_keys: Iterable[AssetKey]
    ) -> Optional[JobDefinition]:
        """Returns the asset base job that contains all the given assets, or None if there is no such
        job.
        """
        if self.has_job(ASSET_BASE_JOB_PREFIX):
            base_job = self.get_job(ASSET_BASE_JOB_PREFIX)
            if all(
                key in base_job.asset_layer.assets_defs_by_key
                or base_job.asset_layer.is_observable_for_asset(key)
                for key in asset_keys
            ):
                return base_job
        else:
            i = 0
            while self.has_job(f"{ASSET_BASE_JOB_PREFIX}_{i}"):
                base_job = self.get_job(f"{ASSET_BASE_JOB_PREFIX}_{i}")

                if all(
                    key in base_job.asset_layer.assets_defs_by_key
                    or base_job.asset_layer.is_observable_for_asset(key)
                    for key in asset_keys
                ):
                    return base_job

                i += 1

        return None

    def get_maybe_subset_job_def(
        self,
        job_name: str,
        op_selection: Optional[Iterable[str]] = None,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
        asset_check_selection: Optional[AbstractSet[AssetCheckKey]] = None,
    ):
        defn = self.get_job(job_name)
        return defn.get_subset(
            op_selection=op_selection,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
        )

    @public
    def load_asset_value(
        self,
        asset_key: CoercibleToAssetKey,
        *,
        python_type: Optional[Type] = None,
        instance: Optional[DagsterInstance] = None,
        partition_key: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        resource_config: Optional[Any] = None,
    ) -> object:
        """Load the contents of an asset as a Python object.

        Invokes `load_input` on the :py:class:`IOManager` associated with the asset.

        If you want to load the values of multiple assets, it's more efficient to use
        :py:meth:`~dagster.RepositoryDefinition.get_asset_value_loader`, which avoids spinning up
        resources separately for each asset.

        Args:
            asset_key (Union[AssetKey, Sequence[str], str]): The key of the asset to load.
            python_type (Optional[Type]): The python type to load the asset as. This is what will
                be returned inside `load_input` by `context.dagster_type.typing_type`.
            partition_key (Optional[str]): The partition of the asset to load.
            metadata (Optional[Dict[str, Any]]): Input metadata to pass to the :py:class:`IOManager`
                (is equivalent to setting the metadata argument in `In` or `AssetIn`).
            resource_config (Optional[Any]): A dictionary of resource configurations to be passed
                to the :py:class:`IOManager`.

        Returns:
            The contents of an asset as a Python object.
        """
        from dagster._core.storage.asset_value_loader import AssetValueLoader

        with AssetValueLoader(
            self.assets_defs_by_key, self.source_assets_by_key, instance=instance
        ) as loader:
            return loader.load_asset_value(
                asset_key,
                python_type=python_type,
                partition_key=partition_key,
                metadata=metadata,
                resource_config=resource_config,
            )

    @public
    def get_asset_value_loader(
        self, instance: Optional[DagsterInstance] = None
    ) -> "AssetValueLoader":
        """Returns an object that can load the contents of assets as Python objects.

        Invokes `load_input` on the :py:class:`IOManager` associated with the assets. Avoids
        spinning up resources separately for each asset.

        Usage:

        .. code-block:: python

            with my_repo.get_asset_value_loader() as loader:
                asset1 = loader.load_asset_value("asset1")
                asset2 = loader.load_asset_value("asset2")

        """
        from dagster._core.storage.asset_value_loader import AssetValueLoader

        return AssetValueLoader(
            self.assets_defs_by_key, self.source_assets_by_key, instance=instance
        )

    @property
    def asset_graph(self) -> InternalAssetGraph:
        return AssetGraph.from_assets(
            [*set(self.assets_defs_by_key.values()), *self.source_assets_by_key.values()]
        )

    # If definition comes from the @repository decorator, then the __call__ method will be
    # overwritten. Therefore, we want to maintain the call-ability of repository definitions.
    def __call__(self, *args, **kwargs):
        return self


class PendingRepositoryDefinition:
    def __init__(
        self,
        name: str,
        repository_definitions: Sequence[
            Union[RepositoryListDefinition, "CacheableAssetsDefinition"]
        ],
        description: Optional[str] = None,
        metadata: Optional[MetadataMapping] = None,
        default_logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
        default_executor_def: Optional[ExecutorDefinition] = None,
        _top_level_resources: Optional[Mapping[str, ResourceDefinition]] = None,
        _resource_key_mapping: Optional[Mapping[int, str]] = None,
    ):
        self._repository_definitions = check.list_param(
            repository_definitions,
            "repository_definition",
            additional_message=(
                "PendingRepositoryDefinition supports only list-based repository data at this time."
            ),
        )
        self._name = name
        self._description = description
        self._metadata = metadata
        self._default_logger_defs = default_logger_defs
        self._default_executor_def = default_executor_def
        self._top_level_resources = _top_level_resources
        self._resource_key_mapping = _resource_key_mapping

    @property
    def name(self) -> str:
        return self._name

    def _compute_repository_load_data(self) -> RepositoryLoadData:
        from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition

        return RepositoryLoadData(
            cached_data_by_key={
                defn.unique_id: defn.compute_cacheable_data()
                for defn in self._repository_definitions
                if isinstance(defn, CacheableAssetsDefinition)
            }
        )

    def _get_repository_definition(
        self, repository_load_data: RepositoryLoadData
    ) -> RepositoryDefinition:
        from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition

        resolved_definitions: List[RepositoryListDefinition] = []
        for defn in self._repository_definitions:
            if isinstance(defn, CacheableAssetsDefinition):
                # should always have metadata for each cached defn at this point
                check.invariant(
                    defn.unique_id in repository_load_data.cached_data_by_key,
                    "No metadata found for CacheableAssetsDefinition with unique_id"
                    f" {defn.unique_id}.",
                )
                # use the emtadata to generate definitions
                resolved_definitions.extend(
                    defn.build_definitions(
                        data=repository_load_data.cached_data_by_key[defn.unique_id]
                    )
                )
            else:
                resolved_definitions.append(defn)

        repository_data = CachingRepositoryData.from_list(
            resolved_definitions,
            default_executor_def=self._default_executor_def,
            default_logger_defs=self._default_logger_defs,
            top_level_resources=self._top_level_resources,
            resource_key_mapping=self._resource_key_mapping,
        )

        return RepositoryDefinition(
            self._name,
            repository_data=repository_data,
            description=self._description,
            metadata=self._metadata,
            repository_load_data=repository_load_data,
        )

    def reconstruct_repository_definition(
        self, repository_load_data: RepositoryLoadData
    ) -> RepositoryDefinition:
        """Use the provided RepositoryLoadData to construct and return a RepositoryDefinition."""
        check.inst_param(repository_load_data, "repository_load_data", RepositoryLoadData)
        return self._get_repository_definition(repository_load_data)

    def compute_repository_definition(self) -> RepositoryDefinition:
        """Compute the required RepositoryLoadData and use it to construct and return a RepositoryDefinition."""
        repository_load_data = self._compute_repository_load_data()
        return self._get_repository_definition(repository_load_data)



import asyncio
import inspect
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem

from dagster import (
    InputContext,
    MetadataValue,
    MultiPartitionKey,
    OutputContext,
    _check as check,
)
from dagster._core.storage.memoizable_io_manager import MemoizableIOManager

if TYPE_CHECKING:
    from upath import UPath


class UPathIOManager(MemoizableIOManager):
    """Abstract IOManager base class compatible with local and cloud storage via `universal-pathlib` and `fsspec`.

    Features:
     - handles partitioned assets
     - handles loading a single upstream partition
     - handles loading multiple upstream partitions (with respect to :py:class:`PartitionMapping`)
     - supports loading multiple partitions concurrently with async `load_from_path` method
     - the `get_metadata` method can be customized to add additional metadata to the output
     - the `allow_missing_partitions` metadata value can be set to `True` to skip missing partitions
       (the default behavior is to raise an error)

    """

    extension: Optional[str] = None  # override in child class

    def __init__(
        self,
        base_path: Optional["UPath"] = None,
    ):
        from upath import UPath

        assert not self.extension or "." in self.extension
        self._base_path = base_path or UPath(".")

    @abstractmethod
    def dump_to_path(self, context: OutputContext, obj: Any, path: "UPath"):
        """Child classes should override this method to write the object to the filesystem."""

    @abstractmethod
    def load_from_path(self, context: InputContext, path: "UPath") -> Any:
        """Child classes should override this method to load the object from the filesystem."""

    @property
    def fs(self) -> AbstractFileSystem:
        """Utility function to get the IOManager filesystem.

        Returns:
            AbstractFileSystem: fsspec filesystem.

        """
        from upath import UPath

        if isinstance(self._base_path, UPath):
            return self._base_path.fs
        elif isinstance(self._base_path, Path):
            return LocalFileSystem()
        else:
            raise ValueError(f"Unsupported base_path type: {type(self._base_path)}")

    @property
    def storage_options(self) -> Dict[str, Any]:
        """Utility function to get the fsspec storage_options which are often consumed by various I/O functions.

        Returns:
            Dict[str, Any]: fsspec storage_options.
        """
        from upath import UPath

        if isinstance(self._base_path, UPath):
            return self._base_path._kwargs.copy()  # noqa
        elif isinstance(self._base_path, Path):
            return {}
        else:
            raise ValueError(f"Unsupported base_path type: {type(self._base_path)}")

    def get_metadata(
        self,
        context: OutputContext,
        obj: Any,
    ) -> Dict[str, MetadataValue]:
        """Child classes should override this method to add custom metadata to the outputs."""
        return {}

    # Read/write operations on paths can generally be handled by methods on the
    # UPath class, but when the backend requires credentials, this isn't
    # always possible. Override these path_* methods to provide custom
    # implementations for targeting backends that require authentication.

    def unlink(self, path: "UPath") -> None:
        """Remove the file or object at the provided path."""
        path.unlink()

    def path_exists(self, path: "UPath") -> bool:
        """Check if a file or object exists at the provided path."""
        return path.exists()

    def make_directory(self, path: "UPath"):
        """Create a directory at the provided path.

        Override as a no-op if the target backend doesn't use directories.
        """
        path.mkdir(parents=True, exist_ok=True)

    def has_output(self, context: OutputContext) -> bool:
        return self.path_exists(self._get_path(context))

    def _with_extension(self, path: "UPath") -> "UPath":
        return path.with_suffix(path.suffix + self.extension) if self.extension else path

    def _get_path_without_extension(self, context: Union[InputContext, OutputContext]) -> "UPath":
        if context.has_asset_key:
            context_path = self.get_asset_relative_path(context)
        else:
            # we are dealing with an op output
            context_path = self.get_op_output_relative_path(context)

        return self._base_path.joinpath(context_path)

    def get_asset_relative_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        from upath import UPath

        # we are not using context.get_asset_identifier() because it already includes the partition_key
        return UPath(*context.asset_key.path)

    def get_op_output_relative_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        from upath import UPath

        return UPath(*context.get_identifier())

    def get_loading_input_log_message(self, path: "UPath") -> str:
        return f"Loading file from: {path} using {self.__class__.__name__}..."

    def get_writing_output_log_message(self, path: "UPath") -> str:
        return f"Writing file at: {path} using {self.__class__.__name__}..."

    def get_loading_input_partition_log_message(self, path: "UPath", partition_key: str) -> str:
        return f"Loading partition {partition_key} from {path} using {self.__class__.__name__}..."

    def get_missing_partition_log_message(self, partition_key: str) -> str:
        return (
            f"Couldn't load partition {partition_key} and skipped it "
            "because the input metadata includes allow_missing_partitions=True"
        )

    def _get_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        """Returns the I/O path for a given context.
        Should not be used with partitions (use `_get_paths_for_partitions` instead).
        """
        path = self._get_path_without_extension(context)
        return self._with_extension(path)

    def get_path_for_partition(
        self, context: Union[InputContext, OutputContext], path: "UPath", partition: str
    ) -> "UPath":
        """Override this method if you want to use a different partitioning scheme
        (for example, if the saving function handles partitioning instead).
        The extension will be added later.

        Args:
            context (Union[InputContext, OutputContext]): The context for the I/O operation.
            path (UPath): The path to the file or object.
            partition (str): Formatted partition/multipartition key

        Returns:
            UPath: The path to the file with the partition key appended.
        """
        return path / partition

    def _get_paths_for_partitions(
        self, context: Union[InputContext, OutputContext]
    ) -> Dict[str, "UPath"]:
        """Returns a dict of partition_keys into I/O paths for a given context."""
        if not context.has_asset_partitions:
            raise TypeError(
                f"Detected {context.dagster_type.typing_type} input type "
                "but the asset is not partitioned"
            )

        def _formatted_multipartitioned_path(partition_key: MultiPartitionKey) -> str:
            ordered_dimension_keys = [
                key[1]
                for key in sorted(partition_key.keys_by_dimension.items(), key=lambda x: x[0])
            ]
            return "/".join(ordered_dimension_keys)

        formatted_partition_keys = {
            partition_key: (
                _formatted_multipartitioned_path(partition_key)
                if isinstance(partition_key, MultiPartitionKey)
                else partition_key
            )
            for partition_key in context.asset_partition_keys
        }

        asset_path = self._get_path_without_extension(context)
        return {
            partition_key: self._with_extension(
                self.get_path_for_partition(context, asset_path, partition)
            )
            for partition_key, partition in formatted_partition_keys.items()
        }

    def _get_multipartition_backcompat_paths(
        self, context: Union[InputContext, OutputContext]
    ) -> Mapping[str, "UPath"]:
        if not context.has_asset_partitions:
            raise TypeError(
                f"Detected {context.dagster_type.typing_type} input type "
                "but the asset is not partitioned"
            )

        partition_keys = context.asset_partition_keys

        asset_path = self._get_path_without_extension(context)
        return {
            partition_key: self._with_extension(asset_path / partition_key)
            for partition_key in partition_keys
            if isinstance(partition_key, MultiPartitionKey)
        }

    def _load_single_input(
        self, path: "UPath", context: InputContext, backcompat_path: Optional["UPath"] = None
    ) -> Any:
        context.log.debug(self.get_loading_input_log_message(path))
        try:
            obj = self.load_from_path(context=context, path=path)
            if asyncio.iscoroutine(obj):
                obj = asyncio.run(obj)
        except FileNotFoundError as e:
            if backcompat_path is not None:
                try:
                    obj = self.load_from_path(context=context, path=backcompat_path)
                    if asyncio.iscoroutine(obj):
                        obj = asyncio.run(obj)

                    context.log.debug(
                        f"File not found at {path}. Loaded instead from backcompat path:"
                        f" {backcompat_path}"
                    )
                except FileNotFoundError:
                    raise e
            else:
                raise e

        context.add_input_metadata({"path": MetadataValue.path(str(path))})
        return obj

    def _load_partition_from_path(
        self,
        context: InputContext,
        partition_key: str,
        path: "UPath",
        backcompat_path: Optional["UPath"] = None,
    ) -> Any:
        """1. Try to load the partition from the normal path.
        2. If it was not found, try to load it from the backcompat path.
        3. If allow_missing_partitions metadata is True, skip the partition if it was not found in any of the paths.
        Otherwise, raise an error.

        Args:
            context (InputContext): IOManager Input context
            partition_key (str): the partition key corresponding to the partition being loaded
            path (UPath): The path to the partition.
            backcompat_path (Optional[UPath]): The path to the partition in the backcompat location.

        Returns:
            Any: The object loaded from the partition.
        """
        allow_missing_partitions = (
            context.metadata.get("allow_missing_partitions", False)
            if context.metadata is not None
            else False
        )

        try:
            context.log.debug(self.get_loading_input_partition_log_message(path, partition_key))
            obj = self.load_from_path(context=context, path=path)
            return obj
        except FileNotFoundError as e:
            if backcompat_path is not None:
                try:
                    obj = self.load_from_path(context=context, path=path)
                    context.log.debug(
                        f"File not found at {path}. Loaded instead from backcompat path:"
                        f" {backcompat_path}"
                    )
                    return obj
                except FileNotFoundError as e:
                    if allow_missing_partitions:
                        context.log.warning(self.get_missing_partition_log_message(partition_key))
                        return None
                    else:
                        raise e
            if allow_missing_partitions:
                context.log.warning(self.get_missing_partition_log_message(partition_key))
                return None
            else:
                raise e

    def _load_multiple_inputs(self, context: InputContext) -> Dict[str, Any]:
        # load multiple partitions
        paths = self._get_paths_for_partitions(context)  # paths for normal partitions
        backcompat_paths = self._get_multipartition_backcompat_paths(
            context
        )  # paths for multipartitions

        context.log.debug(f"Loading {len(paths)} partitions...")

        objs = {}

        if not inspect.iscoroutinefunction(self.load_from_path):
            for partition_key in context.asset_partition_keys:
                obj = self._load_partition_from_path(
                    context,
                    partition_key,
                    paths[partition_key],
                    backcompat_paths.get(partition_key),
                )
                if obj is not None:  # in case some partitions were skipped
                    objs[partition_key] = obj
            return objs
        else:
            # load_from_path returns a coroutine, so we need to await the results

            async def collect():
                loop = asyncio.get_running_loop()

                tasks = []

                for partition_key in context.asset_partition_keys:
                    tasks.append(
                        loop.create_task(
                            self._load_partition_from_path(
                                context,
                                partition_key,
                                paths[partition_key],
                                backcompat_paths.get(partition_key),
                            )
                        )
                    )

                results = await asyncio.gather(*tasks, return_exceptions=True)

                # need to handle missing partitions here because exceptions don't get propagated from async calls
                allow_missing_partitions = (
                    context.metadata.get("allow_missing_partitions", False)
                    if context.metadata is not None
                    else False
                )

                results_without_errors = []
                found_errors = False
                for partition_key, result in zip(context.asset_partition_keys, results):
                    if isinstance(result, FileNotFoundError):
                        if allow_missing_partitions:
                            context.log.warning(
                                self.get_missing_partition_log_message(partition_key)
                            )
                        else:
                            context.log.error(str(result))
                            found_errors = True
                    elif isinstance(result, Exception):
                        context.log.error(str(result))
                        found_errors = True
                    else:
                        results_without_errors.append(result)

                if found_errors:
                    raise RuntimeError(
                        f"{len(paths) - len(results_without_errors)} partitions could not be loaded"
                    )

                return results_without_errors

            awaited_objects = asyncio.get_event_loop().run_until_complete(collect())

            return {
                partition_key: awaited_object
                for partition_key, awaited_object in zip(
                    context.asset_partition_keys, awaited_objects
                )
                if awaited_object is not None
            }

    def load_input(self, context: InputContext) -> Union[Any, Dict[str, Any]]:
        # If no asset key, we are dealing with an op output which is always non-partitioned
        if not context.has_asset_key or not context.has_asset_partitions:
            path = self._get_path(context)
            return self._load_single_input(path, context)
        else:
            asset_partition_keys = context.asset_partition_keys
            if len(asset_partition_keys) == 0:
                return None
            elif len(asset_partition_keys) == 1:
                paths = self._get_paths_for_partitions(context)
                check.invariant(len(paths) == 1, f"Expected 1 path, but got {len(paths)}")
                path = next(iter(paths.values()))
                backcompat_paths = self._get_multipartition_backcompat_paths(context)
                backcompat_path = (
                    None if not backcompat_paths else next(iter(backcompat_paths.values()))
                )

                return self._load_single_input(path, context, backcompat_path)
            else:  # we are dealing with multiple partitions of an asset
                type_annotation = context.dagster_type.typing_type
                if type_annotation != Any and not is_dict_type(type_annotation):
                    check.failed(
                        "Loading an input that corresponds to multiple partitions, but the"
                        " type annotation on the op input is not a dict, Dict, Mapping, or"
                        f" Any: is '{type_annotation}'."
                    )

                return self._load_multiple_inputs(context)

    def handle_output(self, context: OutputContext, obj: Any):
        if context.dagster_type.typing_type == type(None):
            check.invariant(
                obj is None,
                "Output had Nothing type or 'None' annotation, but handle_output received"
                f" value that was not None and was of type {type(obj)}.",
            )
            return None

        if context.has_asset_partitions:
            paths = self._get_paths_for_partitions(context)

            check.invariant(
                len(paths) == 1,
                f"The current IO manager {type(self)} does not support persisting an output"
                " associated with multiple partitions. This error is likely occurring because a"
                " backfill was launched using the 'single run' option. Instead, launch the"
                " backfill with the 'multiple runs' option.",
            )

            path = next(iter(paths.values()))
        else:
            path = self._get_path(context)
        self.make_directory(path.parent)
        context.log.debug(self.get_writing_output_log_message(path))
        self.dump_to_path(context=context, obj=obj, path=path)

        metadata = {"path": MetadataValue.path(str(path))}
        custom_metadata = self.get_metadata(context=context, obj=obj)
        metadata.update(custom_metadata)  # type: ignore

        context.add_output_metadata(metadata)


def is_dict_type(type_obj) -> bool:
    if type_obj == dict:
        return True

    if hasattr(type_obj, "__origin__") and type_obj.__origin__ in (dict, Dict, Mapping):
        return True

    return False



from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey, AssetCheckSpec
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.metadata import (
    ArbitraryMetadataMapping,
    RawMetadataValue,
)
from dagster._core.selector.subset_selector import AssetSelectionData

from ..errors import (
    DagsterInvalidSubsetError,
    DagsterInvariantViolationError,
)
from .config import ConfigMapping
from .dependency import NodeHandle, NodeInputHandle, NodeOutput, NodeOutputHandle
from .events import AssetKey
from .executor_definition import ExecutorDefinition
from .graph_definition import GraphDefinition
from .node_definition import NodeDefinition
from .resource_definition import ResourceDefinition

if TYPE_CHECKING:
    from dagster._core.definitions.assets import AssetsDefinition, SourceAsset
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.partition_mapping import PartitionMapping
    from dagster._core.definitions.resolved_asset_deps import ResolvedAssetDependencies
    from dagster._core.execution.context.output import OutputContext

    from .partition import PartitionedConfig, PartitionsDefinition


class AssetOutputInfo(
    NamedTuple(
        "_AssetOutputInfo",
        [
            ("key", AssetKey),
            ("partitions_fn", Callable[["OutputContext"], Optional[AbstractSet[str]]]),
            ("partitions_def", Optional["PartitionsDefinition"]),
            ("is_required", bool),
            ("code_version", Optional[str]),
        ],
    )
):
    """Defines all of the asset-related information for a given output.

    Args:
        key (AssetKey): The AssetKey
        partitions_fn (OutputContext -> Optional[Set[str]], optional): A function which takes the
            current OutputContext and generates a set of partition names that will be materialized
            for this asset.
        partitions_def (PartitionsDefinition, optional): Defines the set of valid partitions
            for this asset.
        code_version (Optional[str], optional): The version of the code that generates this asset.
    """

    def __new__(
        cls,
        key: AssetKey,
        partitions_fn: Optional[Callable[["OutputContext"], Optional[AbstractSet[str]]]] = None,
        partitions_def: Optional["PartitionsDefinition"] = None,
        is_required: bool = True,
        code_version: Optional[str] = None,
    ):
        return super().__new__(
            cls,
            key=check.inst_param(key, "key", AssetKey),
            partitions_fn=check.opt_callable_param(partitions_fn, "partitions_fn", lambda _: None),
            partitions_def=partitions_def,
            is_required=is_required,
            code_version=code_version,
        )


def _resolve_output_to_destinations(
    output_name: str, node_def: NodeDefinition, handle: NodeHandle
) -> Sequence[NodeInputHandle]:
    all_destinations: List[NodeInputHandle] = []
    if not isinstance(node_def, GraphDefinition):
        # must be in the op definition
        return all_destinations

    for mapping in node_def.output_mappings:
        if mapping.graph_output_name != output_name:
            continue
        output_pointer = mapping.maps_from
        output_node = node_def.node_named(output_pointer.node_name)

        all_destinations.extend(
            _resolve_output_to_destinations(
                output_pointer.output_name,
                output_node.definition,
                NodeHandle(output_pointer.node_name, parent=handle),
            )
        )

        output_def = output_node.definition.output_def_named(output_pointer.output_name)
        downstream_input_handles = (
            node_def.dependency_structure.output_to_downstream_inputs_for_node(
                output_pointer.node_name
            ).get(NodeOutput(output_node, output_def), [])
        )
        for input_handle in downstream_input_handles:
            all_destinations.append(
                NodeInputHandle(
                    NodeHandle(input_handle.node_name, parent=handle), input_handle.input_name
                )
            )

    return all_destinations


def _build_graph_dependencies(
    graph_def: GraphDefinition,
    parent_handle: Optional[NodeHandle],
    outputs_by_graph_handle: Dict[NodeHandle, Mapping[str, NodeOutputHandle]],
    non_asset_inputs_by_node_handle: Dict[NodeHandle, Sequence[NodeOutputHandle]],
    assets_defs_by_node_handle: Mapping[NodeHandle, "AssetsDefinition"],
) -> None:
    """Scans through every node in the graph, making a recursive call when a node is a graph.

    Builds two dictionaries:

    outputs_by_graph_handle: A mapping of every graph node handle to a dictionary with each out
        name as a key and a NodeOutputHandle containing the op output name and op node handle

    non_asset_inputs_by_node_handle: A mapping of all node handles to all upstream node handles
        that are not assets. Each key is a node output handle.
    """
    dep_struct = graph_def.dependency_structure

    for sub_node_name, sub_node in graph_def.node_dict.items():
        curr_node_handle = NodeHandle(sub_node_name, parent=parent_handle)
        if isinstance(sub_node.definition, GraphDefinition):
            _build_graph_dependencies(
                sub_node.definition,
                curr_node_handle,
                outputs_by_graph_handle,
                non_asset_inputs_by_node_handle,
                assets_defs_by_node_handle,
            )
            outputs_by_graph_handle[curr_node_handle] = {
                mapping.graph_output_name: NodeOutputHandle(
                    NodeHandle(mapping.maps_from.node_name, parent=curr_node_handle),
                    mapping.maps_from.output_name,
                )
                for mapping in sub_node.definition.output_mappings
            }
        non_asset_inputs_by_node_handle[curr_node_handle] = [
            NodeOutputHandle(
                NodeHandle(node_output.node_name, parent=parent_handle),
                node_output.output_def.name,
            )
            for node_output in dep_struct.all_upstream_outputs_from_node(sub_node_name)
            if NodeHandle(node_output.node.name, parent=parent_handle)
            not in assets_defs_by_node_handle
        ]


def _get_dependency_node_output_handles(
    non_asset_inputs_by_node_handle: Mapping[NodeHandle, Sequence[NodeOutputHandle]],
    outputs_by_graph_handle: Mapping[NodeHandle, Mapping[str, NodeOutputHandle]],
    dep_node_output_handles_by_node_output_handle: Dict[
        NodeOutputHandle, Sequence[NodeOutputHandle]
    ],
    node_output_handle: NodeOutputHandle,
) -> Sequence[NodeOutputHandle]:
    """Given a node output handle, return all upstream op node output handles. All node output handles
    belong in the same graph-backed asset node.

    Arguments:
    outputs_by_graph_handle: A mapping of every graph node handle to a dictionary with each out
        name as a key and a NodeOutputHandle containing the op output name and op node handle
    non_asset_inputs_by_node_handle: A mapping of all node handles to all upstream node handles
        that are not assets. Each key is a node output handle.
    dep_node_output_handles_by_node_output_handle: A mapping of each non-graph node output handle
        to all non-graph node output handle dependencies. Used for memoization to avoid scanning
        already visited nodes.
    curr_node_handle: The current node handle being traversed.
    graph_output_name: Name of the node output being traversed. Only used if the current node is a
        graph to trace the op that generates this output.
    """
    curr_node_handle = node_output_handle.node_handle

    if node_output_handle in dep_node_output_handles_by_node_output_handle:
        return dep_node_output_handles_by_node_output_handle[node_output_handle]

    dependency_node_output_handles: List[NodeOutputHandle] = (
        []
    )  # first node in list is node output handle that outputs the asset

    if curr_node_handle not in outputs_by_graph_handle:
        dependency_node_output_handles.append(node_output_handle)
    else:  # is graph
        dep_node_output_handle = outputs_by_graph_handle[curr_node_handle][
            node_output_handle.output_name
        ]
        dependency_node_output_handles.extend(
            _get_dependency_node_output_handles(
                non_asset_inputs_by_node_handle,
                outputs_by_graph_handle,
                dep_node_output_handles_by_node_output_handle,
                dep_node_output_handle,
            )
        )
    for dep_node_output_handle in non_asset_inputs_by_node_handle[curr_node_handle]:
        dependency_node_output_handles.extend(
            _get_dependency_node_output_handles(
                non_asset_inputs_by_node_handle,
                outputs_by_graph_handle,
                dep_node_output_handles_by_node_output_handle,
                dep_node_output_handle,
            )
        )

    if curr_node_handle not in outputs_by_graph_handle:
        dep_node_output_handles_by_node_output_handle[node_output_handle] = (
            dependency_node_output_handles
        )

    return dependency_node_output_handles


def get_dep_node_handles_of_graph_backed_asset(
    graph_def: GraphDefinition, assets_def: "AssetsDefinition"
) -> Mapping[AssetKey, Set[NodeHandle]]:
    """Given a graph-backed asset with graph_def, return a mapping of asset keys outputted by the graph
    to a list of node handles within graph_def that are the dependencies of the asset.

    Arguments:
    graph_def: The graph definition of the graph-backed asset.
    assets_def: The assets definition of the graph-backed asset.

    """
    # asset_key_to_dep_node_handles takes in a graph_def that represents the entire job, where each
    # node is a top-level asset node. Create a dummy graph that wraps around graph_def and pass
    # the dummy graph to asset_key_to_dep_node_handles
    dummy_parent_graph = GraphDefinition("dummy_parent_graph", node_defs=[graph_def])
    dep_node_handles_by_asset_key, _ = asset_key_to_dep_node_handles(
        dummy_parent_graph,
        {NodeHandle(name=graph_def.name, parent=None): assets_def},
    )
    return dep_node_handles_by_asset_key


def asset_key_to_dep_node_handles(
    graph_def: GraphDefinition,
    assets_defs_by_node_handle: Mapping[NodeHandle, "AssetsDefinition"],
) -> Tuple[Mapping[AssetKey, Set[NodeHandle]], Mapping[AssetKey, Sequence[NodeOutputHandle]]]:
    """For each asset in assets_defs_by_node_handle, determines all the op handles and output handles
    within the asset's node that are upstream dependencies of the asset.

    Returns a tuple with two objects:
    1. A mapping of each asset key to a set of node handles that are upstream dependencies of the asset.
    2. A mapping of each asset key to a list of node output handles that are upstream dependencies of the asset.

    Arguments:
    graph_def: The graph definition of the job, where each top level node is an asset.
    assets_defs_by_node_handle: A mapping of each node handle to the asset definition for that node.
    """
    # A mapping of all node handles to all upstream node handles
    # that are not assets. Each key is a node handle with node output handle value
    non_asset_inputs_by_node_handle: Dict[NodeHandle, Sequence[NodeOutputHandle]] = {}

    # A mapping of every graph node handle to a dictionary with each out
    # name as a key and node output handle value
    outputs_by_graph_handle: Dict[NodeHandle, Mapping[str, NodeOutputHandle]] = {}
    _build_graph_dependencies(
        graph_def=graph_def,
        parent_handle=None,
        outputs_by_graph_handle=outputs_by_graph_handle,
        non_asset_inputs_by_node_handle=non_asset_inputs_by_node_handle,
        assets_defs_by_node_handle=assets_defs_by_node_handle,
    )

    dep_nodes_by_asset_key: Dict[AssetKey, List[NodeHandle]] = {}
    dep_node_outputs_by_asset_key: Dict[AssetKey, List[NodeOutputHandle]] = {}

    for node_handle, assets_defs in assets_defs_by_node_handle.items():
        dep_node_output_handles_by_node: Dict[NodeOutputHandle, Sequence[NodeOutputHandle]] = (
            {}
        )  # memoized map of node output handles to all node output handle dependencies that are from ops
        for output_name, asset_key in assets_defs.keys_by_output_name.items():
            dep_nodes_by_asset_key[asset_key] = (
                []
            )  # first element in list is node that outputs asset

            dep_node_outputs_by_asset_key[asset_key] = []

            if node_handle not in outputs_by_graph_handle:
                dep_nodes_by_asset_key[asset_key].extend([node_handle])
            else:  # is graph
                # node output handle for the given asset key
                node_output_handle = outputs_by_graph_handle[node_handle][output_name]

                dep_node_output_handles = _get_dependency_node_output_handles(
                    non_asset_inputs_by_node_handle,
                    outputs_by_graph_handle,
                    dep_node_output_handles_by_node,
                    node_output_handle,
                )

                dep_node_outputs_by_asset_key[asset_key].extend(dep_node_output_handles)

    # handle internal_asset_deps within graph-backed assets
    for assets_def in assets_defs_by_node_handle.values():
        for asset_key, dep_asset_keys in assets_def.asset_deps.items():
            if asset_key not in assets_def.keys:
                continue
            for dep_asset_key in [key for key in dep_asset_keys if key in assets_def.keys]:
                if len(dep_node_outputs_by_asset_key[asset_key]) == 0:
                    # This case occurs when the asset is not yielded from a graph-backed asset
                    continue
                node_output_handle = dep_node_outputs_by_asset_key[asset_key][
                    0
                ]  # first item in list is the original node output handle that outputs the asset
                dep_asset_key_node_output_handles = [
                    output_handle
                    for output_handle in dep_node_outputs_by_asset_key[dep_asset_key]
                    if output_handle != node_output_handle
                ]
                dep_node_outputs_by_asset_key[asset_key] = [
                    node_output
                    for node_output in dep_node_outputs_by_asset_key[asset_key]
                    if node_output not in dep_asset_key_node_output_handles
                ]

    # For graph-backed assets, we've resolved the upstream node output handles dependencies for each
    # node output handle in dep_node_outputs_by_asset_key. We use this to find the upstream
    # node handle dependencies.
    for asset_key, dep_node_outputs in dep_node_outputs_by_asset_key.items():
        dep_nodes_by_asset_key[asset_key].extend(
            [node_output.node_handle for node_output in dep_node_outputs]
        )

    dep_node_set_by_asset_key: Dict[AssetKey, Set[NodeHandle]] = {}
    for asset_key, dep_node_handles in dep_nodes_by_asset_key.items():
        dep_node_set_by_asset_key[asset_key] = set(dep_node_handles)
    return dep_node_set_by_asset_key, dep_node_outputs_by_asset_key


class AssetLayer(NamedTuple):
    """Stores all of the asset-related information for a Dagster job. Maps each
    input / output in the underlying graph to the asset it represents (if any), and records the
    dependencies between each asset.

    Args:
        asset_key_by_node_input_handle (Mapping[NodeInputHandle, AssetOutputInfo], optional): A mapping
            from a unique input in the underlying graph to the associated AssetKey that it loads from.
        asset_info_by_node_output_handle (Mapping[NodeOutputHandle, AssetOutputInfo], optional): A mapping
            from a unique output in the underlying graph to the associated AssetOutputInfo.
        asset_deps (Mapping[AssetKey, AbstractSet[AssetKey]], optional): Records the upstream asset
            keys for each asset key produced by this job.
    """

    assets_defs_by_key: Mapping[AssetKey, "AssetsDefinition"]
    assets_defs_by_node_handle: Mapping[NodeHandle, "AssetsDefinition"]
    asset_keys_by_node_input_handle: Mapping[NodeInputHandle, AssetKey]
    asset_info_by_node_output_handle: Mapping[NodeOutputHandle, AssetOutputInfo]
    check_key_by_node_output_handle: Mapping[NodeOutputHandle, AssetCheckKey]
    asset_deps: Mapping[AssetKey, AbstractSet[AssetKey]]
    dependency_node_handles_by_asset_key: Mapping[AssetKey, Set[NodeHandle]]
    source_assets_by_key: Mapping[AssetKey, "SourceAsset"]
    io_manager_keys_by_asset_key: Mapping[AssetKey, str]
    # Used to store the asset key dependencies of op node handles within graph backed assets
    # See AssetLayer.downstream_dep_assets for more information
    dep_asset_keys_by_node_output_handle: Mapping[NodeOutputHandle, Set[AssetKey]]
    partition_mappings_by_asset_dep: Mapping[Tuple[NodeHandle, AssetKey], "PartitionMapping"]
    asset_checks_defs_by_node_handle: Mapping[NodeHandle, "AssetChecksDefinition"]
    node_output_handles_by_asset_check_key: Mapping[AssetCheckKey, NodeOutputHandle]
    check_names_by_asset_key_by_node_handle: Mapping[
        NodeHandle, Mapping[AssetKey, AbstractSet[str]]
    ]

    @staticmethod
    def from_graph_and_assets_node_mapping(
        graph_def: GraphDefinition,
        assets_defs_by_outer_node_handle: Mapping[NodeHandle, "AssetsDefinition"],
        asset_checks_defs_by_node_handle: Mapping[NodeHandle, "AssetChecksDefinition"],
        observable_source_assets_by_node_handle: Mapping[NodeHandle, "SourceAsset"],
        source_assets: Sequence["SourceAsset"],
        resolved_asset_deps: "ResolvedAssetDependencies",
    ) -> "AssetLayer":
        """Generate asset info from a GraphDefinition and a mapping from nodes in that graph to the
        corresponding AssetsDefinition objects.

        Args:
            graph_def (GraphDefinition): The graph for the JobDefinition that we're generating
                this AssetLayer for.
            assets_defs_by_outer_node_handle (Mapping[NodeHandle, AssetsDefinition]): A mapping from
                a NodeHandle pointing to the node in the graph where the AssetsDefinition ended up.
        """
        asset_key_by_input: Dict[NodeInputHandle, AssetKey] = {}
        asset_info_by_output: Dict[NodeOutputHandle, AssetOutputInfo] = {}
        check_key_by_output: Dict[NodeOutputHandle, AssetCheckKey] = {}
        asset_deps: Dict[AssetKey, AbstractSet[AssetKey]] = {}
        io_manager_by_asset: Dict[AssetKey, str] = {
            source_asset.key: source_asset.get_io_manager_key() for source_asset in source_assets
        }
        partition_mappings_by_asset_dep: Dict[Tuple[NodeHandle, AssetKey], "PartitionMapping"] = {}

        (
            dep_node_handles_by_asset_key,
            dep_node_output_handles_by_asset_key,
        ) = asset_key_to_dep_node_handles(graph_def, assets_defs_by_outer_node_handle)

        node_output_handles_by_asset_check_key: Mapping[AssetCheckKey, NodeOutputHandle] = {}
        check_names_by_asset_key_by_node_handle: Dict[NodeHandle, Dict[AssetKey, Set[str]]] = {}

        for node_handle, assets_def in assets_defs_by_outer_node_handle.items():
            for key in assets_def.keys:
                asset_deps[key] = resolved_asset_deps.get_resolved_upstream_asset_keys(
                    assets_def, key
                )

            for input_name in assets_def.node_keys_by_input_name.keys():
                resolved_asset_key = resolved_asset_deps.get_resolved_asset_key_for_input(
                    assets_def, input_name
                )
                input_handle = NodeInputHandle(node_handle, input_name)
                asset_key_by_input[input_handle] = resolved_asset_key
                # resolve graph input to list of op inputs that consume it
                node_input_handles = assets_def.node_def.resolve_input_to_destinations(input_handle)
                for node_input_handle in node_input_handles:
                    asset_key_by_input[node_input_handle] = resolved_asset_key

                partition_mapping = assets_def.get_partition_mapping_for_input(input_name)
                if partition_mapping is not None:
                    partition_mappings_by_asset_dep[(node_handle, resolved_asset_key)] = (
                        partition_mapping
                    )

            for output_name, asset_key in assets_def.node_keys_by_output_name.items():
                # resolve graph output to the op output it comes from
                inner_output_def, inner_node_handle = assets_def.node_def.resolve_output_to_origin(
                    output_name, handle=node_handle
                )
                node_output_handle = NodeOutputHandle(
                    check.not_none(inner_node_handle), inner_output_def.name
                )

                def partitions_fn(context: "OutputContext") -> AbstractSet[str]:
                    from dagster._core.definitions.partition import PartitionsDefinition

                    if context.has_partition_key:
                        return {context.partition_key}

                    return set(
                        cast(
                            PartitionsDefinition, context.asset_partitions_def
                        ).get_partition_keys_in_range(
                            context.asset_partition_key_range,
                            dynamic_partitions_store=context.step_context.instance,
                        )
                    )

                asset_info_by_output[node_output_handle] = AssetOutputInfo(
                    asset_key,
                    partitions_fn=partitions_fn if assets_def.partitions_def else None,
                    partitions_def=assets_def.partitions_def,
                    is_required=asset_key in assets_def.keys,
                    code_version=inner_output_def.code_version,
                )
                io_manager_by_asset[asset_key] = inner_output_def.io_manager_key

                asset_key_by_input.update(
                    {
                        input_handle: asset_key
                        for input_handle in _resolve_output_to_destinations(
                            output_name, assets_def.node_def, node_handle
                        )
                    }
                )

            if len(assets_def.check_specs_by_output_name) > 0:
                check_names_by_asset_key_by_node_handle[node_handle] = defaultdict(set)

                for output_name, check_spec in assets_def.check_specs_by_output_name.items():
                    inner_output_def, inner_node_handle = (
                        assets_def.node_def.resolve_output_to_origin(
                            output_name, handle=node_handle
                        )
                    )
                    node_output_handle = NodeOutputHandle(
                        check.not_none(inner_node_handle), inner_output_def.name
                    )
                    node_output_handles_by_asset_check_key[check_spec.key] = node_output_handle
                    check_names_by_asset_key_by_node_handle[node_handle][check_spec.asset_key].add(
                        check_spec.name
                    )
                    check_key_by_output[node_output_handle] = check_spec.key

        dep_asset_keys_by_node_output_handle = defaultdict(set)
        for asset_key, node_output_handles in dep_node_output_handles_by_asset_key.items():
            for node_output_handle in node_output_handles:
                dep_asset_keys_by_node_output_handle[node_output_handle].add(asset_key)

        for node_handle, checks_def in asset_checks_defs_by_node_handle.items():
            check_names_by_asset_key_by_node_handle[node_handle] = defaultdict(set)
            for output_name, check_spec in checks_def.specs_by_output_name.items():
                inner_output_def, inner_node_handle = checks_def.node_def.resolve_output_to_origin(
                    output_name, handle=node_handle
                )
                node_output_handle = NodeOutputHandle(
                    check.not_none(inner_node_handle), inner_output_def.name
                )
                node_output_handles_by_asset_check_key[check_spec.key] = node_output_handle
                check_names_by_asset_key_by_node_handle[node_handle][check_spec.asset_key].add(
                    check_spec.name
                )
                check_key_by_output[node_output_handle] = check_spec.key

            for input_name, asset_key in checks_def.asset_keys_by_input_name.items():
                input_handle = NodeInputHandle(node_handle, input_name)
                asset_key_by_input[input_handle] = asset_key
                # resolve graph input to list of op inputs that consume it
                node_input_handles = checks_def.node_def.resolve_input_to_destinations(input_handle)
                for node_input_handle in node_input_handles:
                    asset_key_by_input[node_input_handle] = asset_key

        assets_defs_by_key = {
            key: assets_def
            for assets_def in assets_defs_by_outer_node_handle.values()
            for key in assets_def.keys
        }

        source_assets_by_key = {source_asset.key: source_asset for source_asset in source_assets}
        for node_handle, source_asset in observable_source_assets_by_node_handle.items():
            node_def = cast(NodeDefinition, source_asset.node_def)
            check.invariant(len(node_def.output_defs) == 1)
            output_name = node_def.output_defs[0].name
            # resolve graph output to the op output it comes from
            inner_output_def, inner_node_handle = node_def.resolve_output_to_origin(
                output_name, handle=node_handle
            )
            node_output_handle = NodeOutputHandle(
                check.not_none(inner_node_handle), inner_output_def.name
            )

            asset_info_by_output[node_output_handle] = AssetOutputInfo(
                source_asset.key,
                partitions_fn=None,
                partitions_def=source_asset.partitions_def,
                is_required=True,
                code_version=inner_output_def.code_version,
            )

        assets_defs_by_node_handle: Dict[NodeHandle, "AssetsDefinition"] = {
            # nodes for assets
            **{
                node_handle: assets_defs_by_key[asset_key]
                for asset_key, node_handles in dep_node_handles_by_asset_key.items()
                for node_handle in node_handles
            },
            # nodes for asset checks. Required for AssetsDefs that have selected checks
            # but not assets
            **{
                node_handle: assets_def
                for node_handle, assets_def in assets_defs_by_outer_node_handle.items()
                if assets_def.check_keys
            },
        }

        return AssetLayer(
            asset_keys_by_node_input_handle=asset_key_by_input,
            asset_info_by_node_output_handle=asset_info_by_output,
            check_key_by_node_output_handle=check_key_by_output,
            asset_deps=asset_deps,
            assets_defs_by_node_handle=assets_defs_by_node_handle,
            dependency_node_handles_by_asset_key=dep_node_handles_by_asset_key,
            assets_defs_by_key=assets_defs_by_key,
            source_assets_by_key=source_assets_by_key,
            io_manager_keys_by_asset_key=io_manager_by_asset,
            dep_asset_keys_by_node_output_handle=dep_asset_keys_by_node_output_handle,
            partition_mappings_by_asset_dep=partition_mappings_by_asset_dep,
            asset_checks_defs_by_node_handle=asset_checks_defs_by_node_handle,
            node_output_handles_by_asset_check_key=node_output_handles_by_asset_check_key,
            check_names_by_asset_key_by_node_handle=check_names_by_asset_key_by_node_handle,
        )

    def upstream_assets_for_asset(self, asset_key: AssetKey) -> AbstractSet[AssetKey]:
        check.invariant(
            asset_key in self.asset_deps,
            f"AssetKey '{asset_key}' is not produced by this JobDefinition.",
        )
        return self.asset_deps[asset_key]

    def downstream_assets_for_asset(self, asset_key: AssetKey) -> AbstractSet[AssetKey]:
        return {k for k, v in self.asset_deps.items() if asset_key in v}

    @property
    def asset_keys(self) -> Iterable[AssetKey]:
        return self.dependency_node_handles_by_asset_key.keys()

    @property
    def has_assets_defs(self) -> bool:
        return len(self.assets_defs_by_key) > 0

    @property
    def has_asset_check_defs(self) -> bool:
        return len(self.asset_checks_defs_by_node_handle) > 0

    def has_assets_def_for_asset(self, asset_key: AssetKey) -> bool:
        return asset_key in self.assets_defs_by_key

    def assets_def_for_asset(self, asset_key: AssetKey) -> "AssetsDefinition":
        return self.assets_defs_by_key[asset_key]

    def node_output_handle_for_asset(self, asset_key: AssetKey) -> NodeOutputHandle:
        matching_handles = [
            handle
            for handle, asset_info in self.asset_info_by_node_output_handle.items()
            if asset_info.key == asset_key
        ]
        check.invariant(len(matching_handles) == 1)
        return matching_handles[0]

    def assets_def_for_node(self, node_handle: NodeHandle) -> Optional["AssetsDefinition"]:
        return self.assets_defs_by_node_handle.get(node_handle)

    def asset_key_for_node(self, node_handle: NodeHandle) -> AssetKey:
        assets_def = self.assets_def_for_node(node_handle)
        if not assets_def or len(assets_def.keys_by_output_name.keys()) > 1:
            raise DagsterInvariantViolationError(
                "Cannot call `asset_key_for_node` in a multi_asset with more than one asset."
                " Multiple asset keys defined."
            )
        return next(iter(assets_def.keys_by_output_name.values()))

    def asset_check_specs_for_node(self, node_handle: NodeHandle) -> Sequence[AssetCheckSpec]:
        assets_def_for_node = self.assets_def_for_node(node_handle)
        checks_def_for_node = self.asset_checks_def_for_node(node_handle)

        if assets_def_for_node is not None:
            check.invariant(checks_def_for_node is None)
            return list(assets_def_for_node.check_specs)
        elif checks_def_for_node is not None:
            return list(checks_def_for_node.specs)
        else:
            return []

    def get_spec_for_asset_check(
        self, node_handle: NodeHandle, asset_check_key: AssetCheckKey
    ) -> Optional[AssetCheckSpec]:
        asset_checks_def_or_assets_def = self.asset_checks_defs_by_node_handle.get(
            node_handle
        ) or self.assets_defs_by_node_handle.get(node_handle)
        return (
            asset_checks_def_or_assets_def.get_spec_for_check_key(asset_check_key)
            if asset_checks_def_or_assets_def
            else None
        )

    def get_check_names_by_asset_key_for_node_handle(
        self, node_handle: NodeHandle
    ) -> Mapping[AssetKey, AbstractSet[str]]:
        return self.check_names_by_asset_key_by_node_handle[node_handle]

    def asset_checks_def_for_node(
        self, node_handle: NodeHandle
    ) -> Optional["AssetChecksDefinition"]:
        return self.asset_checks_defs_by_node_handle.get(node_handle)

    @property
    def asset_checks_defs(self) -> Iterable[AssetChecksDefinition]:
        return self.asset_checks_defs_by_node_handle.values()

    def get_output_name_for_asset_check(self, asset_check_key: AssetCheckKey) -> str:
        """Output name in the leaf op."""
        return self.node_output_handles_by_asset_check_key[asset_check_key].output_name

    def asset_key_for_input(self, node_handle: NodeHandle, input_name: str) -> Optional[AssetKey]:
        return self.asset_keys_by_node_input_handle.get(NodeInputHandle(node_handle, input_name))

    def input_for_asset_key(self, node_handle: NodeHandle, key: AssetKey) -> Optional[str]:
        return next(
            (
                input_handle.input_name
                for input_handle, k in self.asset_keys_by_node_input_handle.items()
                if k == key
            ),
            None,
        )

    def io_manager_key_for_asset(self, asset_key: AssetKey) -> str:
        return self.io_manager_keys_by_asset_key.get(asset_key, "io_manager")

    def is_observable_for_asset(self, asset_key: AssetKey) -> bool:
        return (
            asset_key in self.source_assets_by_key
            and self.source_assets_by_key[asset_key].is_observable
        )

    def is_materializable_for_asset(self, asset_key: AssetKey) -> bool:
        return asset_key in self.assets_defs_by_key

    def is_graph_backed_asset(self, asset_key: AssetKey) -> bool:
        assets_def = self.assets_defs_by_key.get(asset_key)
        return False if assets_def is None else isinstance(assets_def.node_def, GraphDefinition)

    def code_version_for_asset(self, asset_key: AssetKey) -> Optional[str]:
        assets_def = self.assets_defs_by_key.get(asset_key)
        if assets_def is not None:
            return assets_def.code_versions_by_key[asset_key]
        else:
            return None

    def metadata_for_asset(
        self, asset_key: AssetKey
    ) -> Optional[Mapping[str, ArbitraryMetadataMapping]]:
        if asset_key in self.source_assets_by_key:
            raw_metadata = self.source_assets_by_key[asset_key].raw_metadata
            return raw_metadata or None
        elif asset_key in self.assets_defs_by_key:
            return self.assets_defs_by_key[asset_key].metadata_by_key[asset_key]
        else:
            check.failed(f"Couldn't find key {asset_key}")

    def asset_info_for_output(
        self, node_handle: NodeHandle, output_name: str
    ) -> Optional[AssetOutputInfo]:
        return self.asset_info_by_node_output_handle.get(NodeOutputHandle(node_handle, output_name))

    def asset_key_for_output(self, node_handle: NodeHandle, output_name: str) -> Optional[AssetKey]:
        asset_info = self.asset_info_for_output(node_handle, output_name)
        if asset_info:
            return asset_info.key
        else:
            return None

    def asset_check_key_for_output(
        self, node_handle: NodeHandle, output_name: str
    ) -> Optional[AssetCheckKey]:
        return self.check_key_by_node_output_handle.get(NodeOutputHandle(node_handle, output_name))

    def group_names_by_assets(self) -> Mapping[AssetKey, str]:
        group_names: Dict[AssetKey, str] = {
            key: assets_def.group_names_by_key[key]
            for key, assets_def in self.assets_defs_by_key.items()
            if key in assets_def.group_names_by_key
        }

        group_names.update(
            {
                key: source_asset_def.group_name
                for key, source_asset_def in self.source_assets_by_key.items()
            }
        )

        return group_names

    def partitions_def_for_asset(self, asset_key: AssetKey) -> Optional["PartitionsDefinition"]:
        assets_def = self.assets_defs_by_key.get(asset_key)

        if assets_def is not None:
            return assets_def.partitions_def
        else:
            source_asset = self.source_assets_by_key.get(asset_key)
            if source_asset is not None:
                return source_asset.partitions_def

        return None

    def partition_mapping_for_node_input(
        self, node_handle: NodeHandle, upstream_asset_key: AssetKey
    ) -> Optional["PartitionMapping"]:
        return self.partition_mappings_by_asset_dep.get((node_handle.root, upstream_asset_key))

    def downstream_dep_assets(self, node_handle: NodeHandle, output_name: str) -> Set[AssetKey]:
        """Given the node handle of an op within a graph-backed asset and an output name,
        returns the asset keys dependent on that output.

        For example, for the following asset:

        @op(out={"out_1": Out(is_required=False), "out_2": Out(is_required=False)})
        def two_outputs_op(context):
            return 1, 1


        @op
        def add_one(x):
            return x + 1


        @graph(out={"asset_one": GraphOut(), "asset_two": GraphOut()})
        def my_graph():
            out_1, out_2 = two_outputs_op()
            return {"asset_one": out_1, "asset_two": add_one(out_2)}

        two_assets = AssetsDefinition.from_graph(my_graph)

        Calling downstream_dep_assets with the node handle of two_outputs_op will return:
        - {AssetKey("asset_one")} if output_name="out_1"
        - {AssetKey("asset_two")} if output_name="out_2"

        Calling downstream_dep_assets with node handle add_one will return:
        - {AssetKey("asset_two")} if output_name="result"
        """
        return self.dep_asset_keys_by_node_output_handle.get(
            NodeOutputHandle(node_handle, output_name), set()
        )


def build_asset_selection_job(
    name: str,
    assets: Iterable["AssetsDefinition"],
    source_assets: Iterable["SourceAsset"],
    asset_checks: Iterable["AssetChecksDefinition"],
    executor_def: Optional[ExecutorDefinition] = None,
    config: Optional[Union[ConfigMapping, Mapping[str, Any], "PartitionedConfig"]] = None,
    partitions_def: Optional["PartitionsDefinition"] = None,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    description: Optional[str] = None,
    tags: Optional[Mapping[str, Any]] = None,
    metadata: Optional[Mapping[str, RawMetadataValue]] = None,
    asset_selection: Optional[AbstractSet[AssetKey]] = None,
    asset_check_selection: Optional[AbstractSet[AssetCheckKey]] = None,
    asset_selection_data: Optional[AssetSelectionData] = None,
    hooks: Optional[AbstractSet[HookDefinition]] = None,
) -> "JobDefinition":
    from dagster._core.definitions.assets_job import build_assets_job

    if asset_selection is None and asset_check_selection is None:
        # no selections, include everything
        included_assets = list(assets)
        excluded_assets = []
        included_source_assets = list(source_assets)
        included_checks_defs = list(asset_checks)
    else:
        # Filter to assets that match either selected assets or include a selected check.
        # E.g. a multi asset can be included even if it's not in asset_selection, if it has a selected check
        # defined with check_specs
        (included_assets, excluded_assets) = _subset_assets_defs(
            assets, asset_selection or set(), asset_check_selection
        )
        included_source_assets = _subset_source_assets(source_assets, asset_selection or set())

        if asset_check_selection is None:
            # If assets were selected and checks are None, then include all checks on the selected assets.
            # Note: once we start explicitly passing in asset checks instead of None from the front end,
            # we can remove this logic.
            included_checks_defs = [
                asset_check
                for asset_check in asset_checks
                if asset_check.asset_key in check.not_none(asset_selection)
            ]
        else:
            # Otherwise, filter to explicitly selected checks defs
            included_checks_defs = [
                asset_check
                for asset_check in asset_checks
                if [spec for spec in asset_check.specs if spec.key in asset_check_selection]
            ]

    if partitions_def:
        for asset in included_assets:
            check.invariant(
                asset.partitions_def == partitions_def or asset.partitions_def is None,
                f"Assets defined for node '{asset.node_def.name}' have a partitions_def of "
                f"{asset.partitions_def}, but job '{name}' has non-matching partitions_def of "
                f"{partitions_def}.",
            )

    if len(included_assets) or len(included_checks_defs) > 0:
        # Job materializes assets and/or executes checks
        final_assets = included_assets
        final_asset_checks = included_checks_defs
        final_source_assets = [*source_assets, *excluded_assets]
    else:
        # Job only observes source assets
        final_assets = []
        final_asset_checks = []
        final_source_assets = included_source_assets

    return build_assets_job(
        name=name,
        assets=final_assets,
        asset_checks=final_asset_checks,
        config=config,
        source_assets=final_source_assets,
        resource_defs=resource_defs,
        executor_def=executor_def,
        partitions_def=partitions_def,
        description=description,
        tags=tags,
        metadata=metadata,
        hooks=hooks,
        _asset_selection_data=asset_selection_data,
    )


def _subset_assets_defs(
    assets: Iterable["AssetsDefinition"],
    selected_asset_keys: AbstractSet[AssetKey],
    selected_asset_check_keys: Optional[AbstractSet[AssetCheckKey]],
) -> Tuple[Sequence["AssetsDefinition"], Sequence["AssetsDefinition"],]:
    """Given a list of asset key selection queries, generate a set of AssetsDefinition objects
    representing the included/excluded definitions.
    """
    included_assets: Set[AssetsDefinition] = set()
    excluded_assets: Set[AssetsDefinition] = set()

    for asset in set(assets):
        # intersection
        selected_subset = selected_asset_keys & asset.keys

        # if specific checks were selected, only include those
        if selected_asset_check_keys is not None:
            selected_check_subset = selected_asset_check_keys & asset.check_keys
        # if no checks were selected, filter to checks that target selected assets
        else:
            selected_check_subset = {
                handle for handle in asset.check_keys if handle.asset_key in selected_subset
            }

        # all assets in this def are selected
        if selected_subset == asset.keys and selected_check_subset == asset.check_keys:
            included_assets.add(asset)
        # no assets in this def are selected
        elif len(selected_subset) == 0 and len(selected_check_subset) == 0:
            excluded_assets.add(asset)
        elif asset.can_subset:
            # subset of the asset that we want
            subset_asset = asset.subset_for(selected_asset_keys, selected_check_subset)
            included_assets.add(subset_asset)
            # subset of the asset that we don't want
            excluded_assets.add(
                asset.subset_for(
                    selected_asset_keys=asset.keys - subset_asset.keys,
                    selected_asset_check_keys=(asset.check_keys - subset_asset.check_keys),
                )
            )
        else:
            raise DagsterInvalidSubsetError(
                f"When building job, the AssetsDefinition '{asset.node_def.name}' "
                f"contains asset keys {sorted(list(asset.keys))}, but "
                f"attempted to select only {sorted(list(selected_subset))}. "
                "This AssetsDefinition does not support subsetting. Please select all "
                "asset keys produced by this asset."
            )

    return (
        list(included_assets),
        list(excluded_assets),
    )


def _subset_source_assets(
    source_assets: Iterable["SourceAsset"],
    selected_asset_keys: AbstractSet[AssetKey],
) -> Sequence["SourceAsset"]:
    return [
        source_asset for source_asset in source_assets if source_asset.key in selected_asset_keys
    ]



import logging
from typing import Iterator, Optional

import pendulum

from dagster._core.storage.dagster_run import (
    FINISHED_STATUSES,
    RunsFilter,
)
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._utils import DebugCrashFlags
from dagster._utils.error import SerializableErrorInfo

RUN_BATCH_SIZE = 1000


def execute_concurrency_slots_iteration(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    _debug_crash_flags: Optional[DebugCrashFlags] = None,
) -> Iterator[Optional[SerializableErrorInfo]]:
    instance = workspace_process_context.instance
    timeout_seconds = instance.run_monitoring_settings.get("free_slots_after_run_end_seconds")
    if not timeout_seconds:
        yield
        return

    if not instance.event_log_storage.supports_global_concurrency_limits:
        yield
        return

    run_ids = instance.event_log_storage.get_concurrency_run_ids()
    if not run_ids:
        yield
        return

    now = pendulum.now("UTC")
    run_records = instance.get_run_records(
        filters=RunsFilter(
            run_ids=list(run_ids),
            statuses=FINISHED_STATUSES,
            updated_before=now.subtract(seconds=timeout_seconds),
        ),
        limit=RUN_BATCH_SIZE,
    )
    for run_record in run_records:
        if run_record.end_time + timeout_seconds < now.timestamp():
            freed_slots = instance.event_log_storage.free_concurrency_slots_for_run(
                run_record.dagster_run.run_id
            )
            if freed_slots:
                logger.info(
                    f"Freed {freed_slots} slots for run {run_record.dagster_run.run_id} with status"
                    f" {run_record.dagster_run.status}"
                )
        yield



import os
import sys
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, cast

import pendulum

import dagster._check as check
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.events import DagsterEvent, DagsterEventType, EngineEventData
from dagster._core.execution.context.system import PlanOrchestrationContext
from dagster._core.execution.plan.active import ActiveExecution
from dagster._core.execution.plan.instance_concurrency_context import InstanceConcurrencyContext
from dagster._core.execution.plan.objects import StepFailureData
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.retries import RetryMode
from dagster._core.executor.step_delegating.step_handler.base import StepHandler, StepHandlerContext
from dagster._core.instance import DagsterInstance
from dagster._grpc.types import ExecuteStepArgs
from dagster._utils.error import serializable_error_info_from_exc_info

from ..base import Executor

if TYPE_CHECKING:
    from dagster._core.execution.plan.step import ExecutionStep

DEFAULT_SLEEP_SECONDS = float(
    os.environ.get("DAGSTER_STEP_DELEGATING_EXECUTOR_SLEEP_SECONDS", "1.0")
)


class StepDelegatingExecutor(Executor):
    """This executor tails the event log for events from the steps that it spins up. It also
    sometimes creates its own events - when it does, that event is automatically written to the
    event log. But we wait until we later tail it from the event log database before yielding it,
    to avoid yielding the same event multiple times to callsites.
    """

    def __init__(
        self,
        step_handler: StepHandler,
        retries: RetryMode,
        sleep_seconds: Optional[float] = None,
        check_step_health_interval_seconds: Optional[int] = None,
        max_concurrent: Optional[int] = None,
        tag_concurrency_limits: Optional[List[Dict[str, Any]]] = None,
        should_verify_step: bool = False,
    ):
        self._step_handler = step_handler
        self._retries = retries

        self._max_concurrent = check.opt_int_param(max_concurrent, "max_concurrent")
        self._tag_concurrency_limits = check.opt_list_param(
            tag_concurrency_limits, "tag_concurrency_limits"
        )

        if self._max_concurrent is not None:
            check.invariant(self._max_concurrent > 0, "max_concurrent must be > 0")

        self._sleep_seconds = cast(
            float,
            check.opt_float_param(sleep_seconds, "sleep_seconds", default=DEFAULT_SLEEP_SECONDS),
        )
        self._check_step_health_interval_seconds = cast(
            int,
            check.opt_int_param(
                check_step_health_interval_seconds, "check_step_health_interval_seconds", default=20
            ),
        )
        self._should_verify_step = should_verify_step
        self._event_cursor: Optional[str] = None

    @property
    def retries(self):
        return self._retries

    def _pop_events(self, instance: DagsterInstance, run_id: str) -> Sequence[DagsterEvent]:
        conn = instance.get_records_for_run(
            run_id,
            self._event_cursor,
            of_type=set(DagsterEventType),
        )
        self._event_cursor = conn.cursor
        dagster_events = [
            record.event_log_entry.dagster_event
            for record in conn.records
            if record.event_log_entry.dagster_event
        ]
        return dagster_events

    def _get_step_handler_context(
        self, plan_context, steps, active_execution
    ) -> StepHandlerContext:
        return StepHandlerContext(
            instance=plan_context.plan_data.instance,
            plan_context=plan_context,
            steps=steps,
            execute_step_args=ExecuteStepArgs(
                job_origin=plan_context.reconstructable_job.get_python_origin(),
                run_id=plan_context.dagster_run.run_id,
                step_keys_to_execute=[step.key for step in steps],
                instance_ref=plan_context.plan_data.instance.get_ref(),
                retry_mode=self.retries.for_inner_plan(),
                known_state=active_execution.get_known_state(),
                should_verify_step=self._should_verify_step,
                print_serialized_events=False,
            ),
            dagster_run=plan_context.dagster_run,
        )

    def execute(self, plan_context: PlanOrchestrationContext, execution_plan: ExecutionPlan):
        check.inst_param(plan_context, "plan_context", PlanOrchestrationContext)
        check.inst_param(execution_plan, "execution_plan", ExecutionPlan)

        DagsterEvent.engine_event(
            plan_context,
            f"Starting execution with step handler {self._step_handler.name}.",
            EngineEventData(),
        )
        with InstanceConcurrencyContext(
            plan_context.instance, plan_context.run_id
        ) as instance_concurrency_context:
            with ActiveExecution(
                execution_plan,
                retry_mode=self.retries,
                max_concurrent=self._max_concurrent,
                tag_concurrency_limits=self._tag_concurrency_limits,
                instance_concurrency_context=instance_concurrency_context,
            ) as active_execution:
                running_steps: Dict[str, ExecutionStep] = {}

                if plan_context.resume_from_failure:
                    DagsterEvent.engine_event(
                        plan_context,
                        "Resuming execution from failure",
                        EngineEventData(),
                    )

                    prior_events = self._pop_events(
                        plan_context.instance,
                        plan_context.run_id,
                    )
                    for dagster_event in prior_events:
                        yield dagster_event

                    possibly_in_flight_steps = active_execution.rebuild_from_events(prior_events)
                    for step in possibly_in_flight_steps:
                        step_handler_context = self._get_step_handler_context(
                            plan_context, [step], active_execution
                        )

                        DagsterEvent.engine_event(
                            step_handler_context.get_step_context(step.key),
                            f"Checking on status of in-progress step {step.key} from previous run",
                            EngineEventData(),
                        )

                        should_retry_step = False
                        health_check = None

                        try:
                            health_check = self._step_handler.check_step_health(
                                step_handler_context
                            )
                        except Exception:
                            # For now we assume that an exception indicates that the step should be resumed.
                            # This should probably be a separate should_resume_step method on the step handler.
                            DagsterEvent.engine_event(
                                step_handler_context.get_step_context(step.key),
                                f"Including {step.key} in the new run since it raised an error"
                                " when checking whether it was running",
                                EngineEventData(
                                    error=serializable_error_info_from_exc_info(sys.exc_info())
                                ),
                            )
                            should_retry_step = True
                        else:
                            if not health_check.is_healthy:
                                DagsterEvent.engine_event(
                                    step_handler_context.get_step_context(step.key),
                                    f"Including step {step.key} in the new run since it is not"
                                    f" currently running: {health_check.unhealthy_reason}",
                                )
                                should_retry_step = True

                        if should_retry_step:
                            # health check failed, launch the step
                            list(
                                self._step_handler.launch_step(
                                    self._get_step_handler_context(
                                        plan_context, [step], active_execution
                                    )
                                )
                            )

                        running_steps[step.key] = step

                last_check_step_health_time = pendulum.now("UTC")

                # Order of events is important here. During an interation, we call handle_event, then get_steps_to_execute,
                # then is_complete. get_steps_to_execute updates the state of ActiveExecution, and without it
                # is_complete can return true when we're just between steps.
                while not active_execution.is_complete:
                    if active_execution.check_for_interrupts():
                        active_execution.mark_interrupted()
                        if not plan_context.instance.run_will_resume(plan_context.run_id):
                            DagsterEvent.engine_event(
                                plan_context,
                                "Executor received termination signal, forwarding to steps",
                                EngineEventData.interrupted(list(running_steps.keys())),
                            )
                            for step in running_steps.values():
                                list(
                                    self._step_handler.terminate_step(
                                        self._get_step_handler_context(
                                            plan_context, [step], active_execution
                                        )
                                    )
                                )
                        else:
                            DagsterEvent.engine_event(
                                plan_context,
                                "Executor received termination signal, not forwarding to steps"
                                " because run will be resumed",
                                EngineEventData(
                                    metadata={
                                        "steps_in_flight": MetadataValue.text(
                                            str(running_steps.keys())
                                        )
                                    },
                                ),
                            )

                        return

                    for dagster_event in self._pop_events(
                        plan_context.instance,
                        plan_context.run_id,
                    ):
                        yield dagster_event
                        # STEP_SKIPPED events are only emitted by ActiveExecution, which already handles
                        # and yields them.

                        if dagster_event.is_step_skipped:
                            assert isinstance(dagster_event.step_key, str)
                            active_execution.verify_complete(plan_context, dagster_event.step_key)
                        else:
                            active_execution.handle_event(dagster_event)
                            if (
                                dagster_event.is_step_success
                                or dagster_event.is_step_failure
                                or dagster_event.is_resource_init_failure
                                or dagster_event.is_step_up_for_retry
                            ):
                                assert isinstance(dagster_event.step_key, str)
                                del running_steps[dagster_event.step_key]

                                if not dagster_event.is_step_up_for_retry:
                                    active_execution.verify_complete(
                                        plan_context, dagster_event.step_key
                                    )

                    # process skips from failures or uncovered inputs
                    list(active_execution.plan_events_iterator(plan_context))

                    curr_time = pendulum.now("UTC")
                    if (
                        curr_time - last_check_step_health_time
                    ).total_seconds() >= self._check_step_health_interval_seconds:
                        last_check_step_health_time = curr_time
                        for step in running_steps.values():
                            step_context = plan_context.for_step(step)

                            try:
                                health_check_result = self._step_handler.check_step_health(
                                    self._get_step_handler_context(
                                        plan_context, [step], active_execution
                                    )
                                )
                                if not health_check_result.is_healthy:
                                    DagsterEvent.step_failure_event(
                                        step_context=step_context,
                                        step_failure_data=StepFailureData(
                                            error=None,
                                            user_failure_data=None,
                                        ),
                                        message=(
                                            f"Step {step.key} failed health check:"
                                            f" {health_check_result.unhealthy_reason}"
                                        ),
                                    )
                            except Exception:
                                serializable_error = serializable_error_info_from_exc_info(
                                    sys.exc_info()
                                )
                                # Log a step failure event if there was an error during the health
                                # check
                                DagsterEvent.step_failure_event(
                                    step_context=plan_context.for_step(step),
                                    step_failure_data=StepFailureData(
                                        error=serializable_error,
                                        user_failure_data=None,
                                    ),
                                )

                    if self._max_concurrent is not None:
                        max_steps_to_run = self._max_concurrent - len(running_steps)
                        check.invariant(
                            max_steps_to_run >= 0, "More steps are active than max_concurrent"
                        )
                    else:
                        max_steps_to_run = None  # disables limit

                    # process events from concurrency blocked steps
                    list(active_execution.concurrency_event_iterator(plan_context))

                    for step in active_execution.get_steps_to_execute(max_steps_to_run):
                        running_steps[step.key] = step
                        list(
                            self._step_handler.launch_step(
                                self._get_step_handler_context(
                                    plan_context, [step], active_execution
                                )
                            )
                        )

                    time.sleep(self._sleep_seconds)



from .asset_check_result import (
    AssetCheckEvaluation as AssetCheckEvaluation,
    AssetCheckResult as AssetCheckResult,
)
from .composition import PendingNodeInvocation as PendingNodeInvocation
from .config import ConfigMapping as ConfigMapping
from .dependency import (
    DependencyDefinition as DependencyDefinition,
    MultiDependencyDefinition as MultiDependencyDefinition,
    Node as Node,
    NodeHandle as NodeHandle,
    NodeInput as NodeInput,
    NodeInvocation as NodeInvocation,
    NodeOutput as NodeOutput,
)
from .events import (
    AssetKey as AssetKey,
    AssetMaterialization as AssetMaterialization,
    AssetObservation as AssetObservation,
    DynamicOutput as DynamicOutput,
    ExpectationResult as ExpectationResult,
    Failure as Failure,
    HookExecutionResult as HookExecutionResult,
    Output as Output,
    RetryRequested as RetryRequested,
    TypeCheck as TypeCheck,
)
from .executor_definition import (
    ExecutorDefinition as ExecutorDefinition,
    ExecutorRequirement as ExecutorRequirement,
    executor as executor,
    in_process_executor as in_process_executor,
    multi_or_in_process_executor as multi_or_in_process_executor,
    multiple_process_executor_requirements as multiple_process_executor_requirements,
    multiprocess_executor as multiprocess_executor,
)
from .hook_definition import HookDefinition as HookDefinition
from .input import (
    GraphIn as GraphIn,
    In as In,
    InputDefinition as InputDefinition,
    InputMapping as InputMapping,
)
from .job_base import IJob as IJob
from .logger_definition import (
    LoggerDefinition as LoggerDefinition,
    build_init_logger_context as build_init_logger_context,
    logger as logger,
)
from .metadata import (
    BoolMetadataValue as BoolMetadataValue,
    DagsterAssetMetadataValue as DagsterAssetMetadataValue,
    DagsterRunMetadataValue as DagsterRunMetadataValue,
    FloatMetadataValue as FloatMetadataValue,
    IntMetadataValue as IntMetadataValue,
    JsonMetadataValue as JsonMetadataValue,
    MarkdownMetadataValue as MarkdownMetadataValue,
    MetadataEntry as MetadataEntry,
    MetadataValue as MetadataValue,
    PathMetadataValue as PathMetadataValue,
    PythonArtifactMetadataValue as PythonArtifactMetadataValue,
    TableColumn as TableColumn,
    TableColumnConstraints as TableColumnConstraints,
    TableConstraints as TableConstraints,
    TableMetadataValue as TableMetadataValue,
    TableRecord as TableRecord,
    TableSchema as TableSchema,
    TableSchemaMetadataValue as TableSchemaMetadataValue,
    TextMetadataValue as TextMetadataValue,
    UrlMetadataValue as UrlMetadataValue,
)
from .node_container import create_execution_structure as create_execution_structure
from .node_definition import NodeDefinition as NodeDefinition
from .output import (
    DynamicOut as DynamicOut,
    DynamicOutputDefinition as DynamicOutputDefinition,
    GraphOut as GraphOut,
    Out as Out,
    OutputDefinition as OutputDefinition,
    OutputMapping as OutputMapping,
)
from .reconstruct import (
    ReconstructableJob as ReconstructableJob,
    build_reconstructable_job as build_reconstructable_job,
    reconstructable as reconstructable,
)
from .repository_definition import (
    RepositoryData as RepositoryData,
    RepositoryDefinition as RepositoryDefinition,
)
from .resolved_asset_deps import ResolvedAssetDependencies as ResolvedAssetDependencies
from .resource_definition import (
    ResourceDefinition as ResourceDefinition,
    make_values_resource as make_values_resource,
    resource as resource,
)
from .run_config_schema import (
    RunConfigSchema as RunConfigSchema,
    create_run_config_schema as create_run_config_schema,
)
from .run_request import (
    AddDynamicPartitionsRequest as AddDynamicPartitionsRequest,
    DeleteDynamicPartitionsRequest as DeleteDynamicPartitionsRequest,
    InstigatorType as InstigatorType,
    RunRequest as RunRequest,
    SensorResult as SensorResult,
    SkipReason as SkipReason,
)
from .schedule_definition import (
    DefaultScheduleStatus as DefaultScheduleStatus,
    ScheduleDefinition as ScheduleDefinition,
    ScheduleEvaluationContext as ScheduleEvaluationContext,
)
from .sensor_definition import (
    DefaultSensorStatus as DefaultSensorStatus,
    SensorDefinition as SensorDefinition,
    SensorEvaluationContext as SensorEvaluationContext,
)

# ruff: isort: split
from .asset_in import AssetIn as AssetIn
from .asset_out import AssetOut as AssetOut
from .asset_selection import AssetSelection as AssetSelection
from .assets import AssetsDefinition as AssetsDefinition
from .assets_job import build_assets_job as build_assets_job
from .decorators import (
    asset as asset,
    asset_sensor as asset_sensor,
    config_mapping as config_mapping,
    failure_hook as failure_hook,
    graph as graph,
    hook_decorator as hook_decorator,
    job as job,
    multi_asset as multi_asset,
    op as op,
    repository as repository,
    schedule as schedule,
    sensor as sensor,
    success_hook as success_hook,
)
from .graph_definition import GraphDefinition as GraphDefinition
from .job_definition import JobDefinition as JobDefinition
from .load_assets_from_modules import (
    load_assets_from_current_module as load_assets_from_current_module,
    load_assets_from_modules as load_assets_from_modules,
    load_assets_from_package_module as load_assets_from_package_module,
    load_assets_from_package_name as load_assets_from_package_name,
)
from .materialize import (
    materialize as materialize,
    materialize_to_memory as materialize_to_memory,
)
from .op_definition import OpDefinition as OpDefinition
from .partition import (
    DynamicPartitionsDefinition as DynamicPartitionsDefinition,
    Partition as Partition,
    PartitionedConfig as PartitionedConfig,
    PartitionsDefinition as PartitionsDefinition,
    StaticPartitionsDefinition as StaticPartitionsDefinition,
    dynamic_partitioned_config as dynamic_partitioned_config,
    static_partitioned_config as static_partitioned_config,
)
from .partition_key_range import PartitionKeyRange as PartitionKeyRange
from .partition_mapping import (
    AllPartitionMapping as AllPartitionMapping,
    DimensionPartitionMapping as DimensionPartitionMapping,
    IdentityPartitionMapping as IdentityPartitionMapping,
    LastPartitionMapping as LastPartitionMapping,
    MultiPartitionMapping as MultiPartitionMapping,
    MultiToSingleDimensionPartitionMapping as MultiToSingleDimensionPartitionMapping,
    PartitionMapping as PartitionMapping,
)
from .partitioned_schedule import (
    build_schedule_from_partitioned_job as build_schedule_from_partitioned_job,
)
from .run_status_sensor_definition import (
    RunFailureSensorContext as RunFailureSensorContext,
    RunStatusSensorContext as RunStatusSensorContext,
    RunStatusSensorDefinition as RunStatusSensorDefinition,
    run_failure_sensor as run_failure_sensor,
    run_status_sensor as run_status_sensor,
)
from .source_asset import SourceAsset as SourceAsset
from .time_window_partition_mapping import TimeWindowPartitionMapping as TimeWindowPartitionMapping
from .time_window_partitions import (
    DailyPartitionsDefinition as DailyPartitionsDefinition,
    HourlyPartitionsDefinition as HourlyPartitionsDefinition,
    MonthlyPartitionsDefinition as MonthlyPartitionsDefinition,
    TimeWindow as TimeWindow,
    TimeWindowPartitionsDefinition as TimeWindowPartitionsDefinition,
    WeeklyPartitionsDefinition as WeeklyPartitionsDefinition,
    daily_partitioned_config as daily_partitioned_config,
    hourly_partitioned_config as hourly_partitioned_config,
    monthly_partitioned_config as monthly_partitioned_config,
    weekly_partitioned_config as weekly_partitioned_config,
)



from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, Mapping, Sequence, Tuple, Union

from dagster import _check as check

if TYPE_CHECKING:
    from dagster._core.execution.plan.step import ExecutionStep
    from dagster._core.storage.dagster_run import DagsterRun


class TagConcurrencyLimitsCounter:
    """Helper object that keeps track of when the tag concurrency limits are met."""

    _key_limits: Dict[str, int]
    _key_value_limits: Dict[Tuple[str, str], int]
    _unique_value_limits: Dict[str, int]
    _key_counts: Dict[str, int]
    _key_value_counts: Dict[Tuple[str, str], int]
    _unique_value_counts: Dict[Tuple[str, str], int]

    def __init__(
        self,
        tag_concurrency_limits: Sequence[Mapping[str, Any]],
        in_progress_tagged_items: Sequence[Union["DagsterRun", "ExecutionStep"]],
    ):
        check.opt_list_param(tag_concurrency_limits, "tag_concurrency_limits", of_type=dict)
        check.list_param(in_progress_tagged_items, "in_progress_tagged_items")

        self._key_limits = {}
        self._key_value_limits = {}
        self._unique_value_limits = {}

        for tag_limit in tag_concurrency_limits:
            key = tag_limit["key"]
            value = tag_limit.get("value")
            limit = tag_limit["limit"]

            if isinstance(value, str):
                self._key_value_limits[(key, value)] = limit
            elif not value or not value["applyLimitPerUniqueValue"]:
                self._key_limits[key] = limit
            else:
                self._unique_value_limits[key] = limit

        self._key_counts = defaultdict(lambda: 0)
        self._key_value_counts = defaultdict(lambda: 0)
        self._unique_value_counts = defaultdict(lambda: 0)

        # initialize counters based on current in progress item
        for item in in_progress_tagged_items:
            self.update_counters_with_launched_item(item)

    def is_blocked(self, item: Union["DagsterRun", "ExecutionStep"]) -> bool:
        """True if there are in progress item which are blocking this item based on tag limits."""
        for key, value in item.tags.items():
            if key in self._key_limits and self._key_counts[key] >= self._key_limits[key]:
                return True

            tag_tuple = (key, value)
            if (
                tag_tuple in self._key_value_limits
                and self._key_value_counts[tag_tuple] >= self._key_value_limits[tag_tuple]
            ):
                return True

            if (
                key in self._unique_value_limits
                and self._unique_value_counts[tag_tuple] >= self._unique_value_limits[key]
            ):
                return True

        return False

    def update_counters_with_launched_item(
        self, item: Union["DagsterRun", "ExecutionStep"]
    ) -> None:
        """Add a new in progress item to the counters."""
        for key, value in item.tags.items():
            if key in self._key_limits:
                self._key_counts[key] += 1

            tag_tuple = (key, value)
            if tag_tuple in self._key_value_limits:
                self._key_value_counts[tag_tuple] += 1

            if key in self._unique_value_limits:
                self._unique_value_counts[tag_tuple] += 1



import inspect
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

from pydantic import ConstrainedFloat, ConstrainedInt, ConstrainedStr
from typing_extensions import Annotated, get_args, get_origin

from dagster import (
    Enum as DagsterEnum,
)
from dagster._config.config_type import (
    Array,
    ConfigType,
    Noneable,
)
from dagster._config.post_process import resolve_defaults
from dagster._config.source import BoolSource, IntSource, StringSource
from dagster._config.validate import validate_config
from dagster._core.definitions.definition_config_schema import (
    DefinitionConfigSchema,
)
from dagster._core.errors import (
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvalidPythonicConfigDefinitionError,
)

from .attach_other_object_to_context import (
    IAttachDifferentObjectToOpContext as IAttachDifferentObjectToOpContext,
)

try:
    from functools import cached_property  # type: ignore  # (py37 compat)
except ImportError:

    class cached_property:
        pass


from pydantic.fields import (
    ModelField,
)

import dagster._check as check
from dagster import Field, Selector
from dagster._config.field_utils import (
    FIELD_NO_DEFAULT_PROVIDED,
    Map,
    convert_potential_field,
)

from .pydantic_compat_layer import ModelFieldCompat, model_fields
from .type_check_utils import is_optional, safe_is_subclass


# This is from https://github.com/dagster-io/dagster/pull/11470
def _apply_defaults_to_schema_field(field: Field, additional_default_values: Any) -> Field:
    # This work by validating the top-level config and then
    # just setting it at that top-level field. Config fields
    # can actually take nested values so we only need to set it
    # at a single level

    evr = validate_config(field.config_type, additional_default_values)

    if not evr.success:
        raise DagsterInvalidConfigError(
            "Incorrect values passed to .configured",
            evr.errors,
            additional_default_values,
        )

    if field.default_provided:
        # In the case where there is already a default config value
        # we can apply "additional" defaults by actually invoking
        # the config machinery. Meaning we pass the new_additional_default_values
        # and then resolve the existing defaults over them. This preserves the default
        # values that are not specified in new_additional_default_values and then
        # applies the new value as the default value of the field in question.
        defaults_processed_evr = resolve_defaults(field.config_type, additional_default_values)
        check.invariant(
            defaults_processed_evr.success,
            "Since validation passed, this should always work.",
        )
        default_to_pass = defaults_processed_evr.value
        return copy_with_default(field, default_to_pass)
    else:
        return copy_with_default(field, additional_default_values)


def copy_with_default(old_field: Field, new_config_value: Any) -> Field:
    return Field(
        config=old_field.config_type,
        default_value=new_config_value,
        is_required=False,
        description=old_field.description,
    )


def _curry_config_schema(schema_field: Field, data: Any) -> DefinitionConfigSchema:
    """Return a new config schema configured with the passed in data."""
    return DefinitionConfigSchema(_apply_defaults_to_schema_field(schema_field, data))


TResValue = TypeVar("TResValue")


def _get_inner_field_if_exists(shape_type: Type, field: ModelField) -> Optional[ModelField]:
    """Grabs the inner Pydantic field type for a data structure such as a list or dictionary.

    Returns None for types which have no inner field.
    """
    # See https://github.com/pydantic/pydantic/blob/v1.10.3/pydantic/fields.py#L758 for
    # where sub_fields is set.
    if safe_is_subclass(get_origin(shape_type), list):
        return check.not_none(get_args(shape_type))[0]
    if safe_is_subclass(get_origin(shape_type), dict) or safe_is_subclass(
        get_origin(shape_type), Mapping
    ):
        return check.not_none(get_args(shape_type))[1]
    return None


def _convert_pydantic_field(
    pydantic_field: ModelFieldCompat, model_cls: Optional[Type] = None
) -> Field:
    """Transforms a Pydantic field into a corresponding Dagster config field.


    Args:
        pydantic_field (ModelField): The Pydantic field to convert.
        model_cls (Optional[Type]): The Pydantic model class that the field belongs to. This is
            used for error messages.
    """
    from .config import Config, infer_schema_from_config_class

    if pydantic_field.discriminator:
        return _convert_pydantic_discriminated_union_field(pydantic_field)

    field_type = pydantic_field.annotation
    if safe_is_subclass(field_type, Config):
        inferred_field = infer_schema_from_config_class(
            field_type,
            description=pydantic_field.description,
        )
        return inferred_field
    else:
        if not pydantic_field.is_required() and not is_optional(field_type):
            field_type = Optional[field_type]

        config_type = _config_type_for_type_on_pydantic_field(field_type)

        return Field(
            config=config_type,
            description=pydantic_field.description,
            is_required=pydantic_field.is_required() and not is_optional(field_type),
            default_value=(
                pydantic_field.default
                if pydantic_field.default is not None
                else FIELD_NO_DEFAULT_PROVIDED
            ),
        )


def strip_wrapping_annotated_types(potentially_annotated_type: Any) -> Any:
    """For a type that is wrapped in Annotated, return the unwrapped type. Recursive,
    so it will unwrap nested Annotated types.

    e.g. Annotated[Annotated[List[str], "foo"], "bar] -> List[str]
    """
    while get_origin(potentially_annotated_type) == Annotated:
        potentially_annotated_type = get_args(potentially_annotated_type)[0]
    return potentially_annotated_type


def _config_type_for_type_on_pydantic_field(
    potential_dagster_type: Any,
) -> ConfigType:
    """Generates a Dagster ConfigType from a Pydantic field's Python type.

    Args:
        potential_dagster_type (Any): The Python type of the Pydantic field.
    """
    potential_dagster_type = strip_wrapping_annotated_types(potential_dagster_type)

    # special case pydantic constrained types to their source equivalents
    if safe_is_subclass(potential_dagster_type, ConstrainedStr):
        return StringSource
    # no FloatSource, so we just return float
    elif safe_is_subclass(potential_dagster_type, ConstrainedFloat):
        potential_dagster_type = float
    elif safe_is_subclass(potential_dagster_type, ConstrainedInt):
        return IntSource

    if safe_is_subclass(get_origin(potential_dagster_type), List):
        list_inner_type = get_args(potential_dagster_type)[0]
        return Array(_config_type_for_type_on_pydantic_field(list_inner_type))
    elif is_optional(potential_dagster_type):
        optional_inner_type = next(
            arg for arg in get_args(potential_dagster_type) if arg is not type(None)
        )
        return Noneable(_config_type_for_type_on_pydantic_field(optional_inner_type))
    elif safe_is_subclass(get_origin(potential_dagster_type), Dict) or safe_is_subclass(
        get_origin(potential_dagster_type), Mapping
    ):
        key_type, value_type = get_args(potential_dagster_type)
        return Map(
            key_type,
            _config_type_for_type_on_pydantic_field(value_type),
        )

    from .config import Config, infer_schema_from_config_class

    if safe_is_subclass(potential_dagster_type, Config):
        inferred_field = infer_schema_from_config_class(
            potential_dagster_type,
        )
        return inferred_field.config_type

    if safe_is_subclass(potential_dagster_type, Enum):
        return DagsterEnum.from_python_enum_direct_values(potential_dagster_type)

    # special case raw python literals to their source equivalents
    if potential_dagster_type is str:
        return StringSource
    elif potential_dagster_type is int:
        return IntSource
    elif potential_dagster_type is bool:
        return BoolSource
    else:
        return convert_potential_field(potential_dagster_type).config_type


def _convert_pydantic_discriminated_union_field(pydantic_field: ModelFieldCompat) -> Field:
    """Builds a Selector config field from a Pydantic field which is a discriminated union.

    For example:

    class Cat(Config):
        pet_type: Literal["cat"]
        meows: int

    class Dog(Config):
        pet_type: Literal["dog"]
        barks: float

    class OpConfigWithUnion(Config):
        pet: Union[Cat, Dog] = Field(..., discriminator="pet_type")

    Becomes:

    Shape({
      "pet": Selector({
          "cat": Shape({"meows": Int}),
          "dog": Shape({"barks": Float}),
      })
    })
    """
    from .config import Config, infer_schema_from_config_class

    field_type = pydantic_field.annotation
    discriminator = pydantic_field.discriminator if pydantic_field.discriminator else None

    if not get_origin(field_type) == Union:
        raise DagsterInvalidDefinitionError("Discriminated union must be a Union type.")

    sub_fields = get_args(field_type)
    if not all(issubclass(sub_field, Config) for sub_field in sub_fields):
        raise NotImplementedError("Discriminated unions with non-Config types are not supported.")

    sub_fields_mapping = {}
    if discriminator:
        for sub_field in sub_fields:
            sub_field_annotation = model_fields(sub_field)[discriminator].annotation

            for sub_field_key in get_args(sub_field_annotation):
                sub_fields_mapping[sub_field_key] = sub_field

    # First, we generate a mapping between the various discriminator values and the
    # Dagster config fields that correspond to them. We strip the discriminator key
    # from the fields, since the user should not have to specify it.

    dagster_config_field_mapping = {
        discriminator_value: infer_schema_from_config_class(
            field,
            fields_to_omit=({discriminator} if discriminator else None),
        )
        for discriminator_value, field in sub_fields_mapping.items()
    }

    # We then nest the union fields under a Selector. The keys for the selector
    # are the various discriminator values
    return Field(config=Selector(fields=dagster_config_field_mapping))


def infer_schema_from_config_annotation(model_cls: Any, config_arg_default: Any) -> Field:
    """Parses a structured config class or primitive type and returns a corresponding Dagster config Field."""
    from .config import Config, infer_schema_from_config_class

    if safe_is_subclass(model_cls, Config):
        check.invariant(
            config_arg_default is inspect.Parameter.empty,
            "Cannot provide a default value when using a Config class",
        )
        return infer_schema_from_config_class(model_cls)

    # If were are here config is annotated with a primitive type
    # We do a conversion to a type as if it were a type on a pydantic field
    try:
        inner_config_type = _config_type_for_type_on_pydantic_field(model_cls)
    except (DagsterInvalidDefinitionError, DagsterInvalidConfigDefinitionError):
        raise DagsterInvalidPythonicConfigDefinitionError(
            invalid_type=model_cls, config_class=None, field_name=None
        )
    return Field(
        config=inner_config_type,
        default_value=(
            FIELD_NO_DEFAULT_PROVIDED
            if config_arg_default is inspect.Parameter.empty
            else config_arg_default
        ),
    )



import logging
import sys
from collections import defaultdict
from types import TracebackType
from typing import (
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
)

import pendulum

import dagster._check as check
from dagster._core.definitions.asset_daemon_context import AssetDaemonContext
from dagster._core.definitions.asset_daemon_cursor import AssetDaemonCursor
from dagster._core.definitions.external_asset_graph import ExternalAssetGraph
from dagster._core.definitions.run_request import (
    InstigatorType,
    RunRequest,
)
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.host_representation.external import (
    ExternalExecutionPlan,
    ExternalJob,
)
from dagster._core.instance import DagsterInstance
from dagster._core.scheduler.instigation import (
    InstigatorTick,
    TickData,
    TickStatus,
)
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import (
    ASSET_EVALUATION_ID_TAG,
    AUTO_MATERIALIZE_TAG,
    AUTO_OBSERVE_TAG,
)
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._core.workspace.workspace import IWorkspace
from dagster._daemon.daemon import DaemonIterator, IntervalDaemon
from dagster._utils import hash_collection
from dagster._utils.error import serializable_error_info_from_exc_info

CURSOR_KEY = "ASSET_DAEMON_CURSOR"
ASSET_DAEMON_PAUSED_KEY = "ASSET_DAEMON_PAUSED"

EVALUATIONS_TTL_DAYS = 30

FIXED_AUTO_MATERIALIZATION_ORIGIN_ID = "asset_daemon_origin"
FIXED_AUTO_MATERIALIZATION_SELECTOR_ID = "asset_daemon_selector"
FIXED_AUTO_MATERIALIZATION_INSTIGATOR_NAME = "asset_daemon"


def get_auto_materialize_paused(instance: DagsterInstance) -> bool:
    return (
        instance.daemon_cursor_storage.get_cursor_values({ASSET_DAEMON_PAUSED_KEY}).get(
            ASSET_DAEMON_PAUSED_KEY
        )
        != "false"
    )


def set_auto_materialize_paused(instance: DagsterInstance, paused: bool):
    instance.daemon_cursor_storage.set_cursor_values(
        {ASSET_DAEMON_PAUSED_KEY: "true" if paused else "false"}
    )


def _get_raw_cursor(instance: DagsterInstance) -> Optional[str]:
    return instance.daemon_cursor_storage.get_cursor_values({CURSOR_KEY}).get(CURSOR_KEY)


def get_current_evaluation_id(instance: DagsterInstance) -> Optional[int]:
    raw_cursor = _get_raw_cursor(instance)
    return AssetDaemonCursor.get_evaluation_id_from_serialized(raw_cursor) if raw_cursor else None


class AutoMaterializeLaunchContext:
    def __init__(
        self,
        tick: InstigatorTick,
        instance: DagsterInstance,
        logger: logging.Logger,
        tick_retention_settings,
    ):
        self._tick = tick
        self._logger = logger
        self._instance = instance

        self._purge_settings = defaultdict(set)
        for status, day_offset in tick_retention_settings.items():
            self._purge_settings[day_offset].add(status)

    @property
    def status(self) -> TickStatus:
        return self._tick.status

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    def add_run_info(self, run_id=None):
        self._tick = self._tick.with_run_info(run_id)

    def set_run_requests(self, run_requests: Sequence[RunRequest]):
        self._tick = self._tick.with_run_requests(run_requests)

    def update_state(self, status: TickStatus, **kwargs: object):
        self._tick = self._tick.with_status(status=status, **kwargs)

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type: Type[BaseException],
        exception_value: Exception,
        traceback: TracebackType,
    ) -> None:
        if exception_value and isinstance(exception_value, KeyboardInterrupt):
            return

        # Log the error if the failure wasn't an interrupt or the daemon generator stopping
        if exception_value and not isinstance(exception_value, GeneratorExit):
            error_data = serializable_error_info_from_exc_info(sys.exc_info())
            self.update_state(TickStatus.FAILURE, error=error_data)

        self._write()

        for day_offset, statuses in self._purge_settings.items():
            if day_offset <= 0:
                continue
            self._instance.purge_ticks(
                FIXED_AUTO_MATERIALIZATION_ORIGIN_ID,
                FIXED_AUTO_MATERIALIZATION_SELECTOR_ID,
                before=pendulum.now("UTC").subtract(days=day_offset).timestamp(),
                tick_statuses=list(statuses),
            )

    def _write(self) -> None:
        self._instance.update_tick(self._tick)


class AssetDaemon(IntervalDaemon):
    def __init__(self, interval_seconds: int):
        super().__init__(interval_seconds=interval_seconds)

    @classmethod
    def daemon_type(cls) -> str:
        return "ASSET"

    def run_iteration(
        self,
        workspace_process_context: IWorkspaceProcessContext,
    ) -> DaemonIterator:
        instance = workspace_process_context.instance

        if get_auto_materialize_paused(instance):
            yield
            return

        schedule_storage = check.not_none(
            instance.schedule_storage,
            "Auto materialization requires schedule storage to be configured",
        )

        if not schedule_storage.supports_auto_materialize_asset_evaluations:
            self._logger.warning(
                "Auto materialize evaluations are not getting logged. Run `dagster instance"
                " migrate` to enable."
            )

        evaluation_time = pendulum.now("UTC")

        workspace = workspace_process_context.create_request_context()
        asset_graph = ExternalAssetGraph.from_workspace(workspace)
        target_asset_keys = {
            target_key
            for target_key in asset_graph.materializable_asset_keys
            if asset_graph.get_auto_materialize_policy(target_key) is not None
        }
        num_target_assets = len(target_asset_keys)

        auto_observe_assets = [
            key
            for key in asset_graph.source_asset_keys
            if asset_graph.get_auto_observe_interval_minutes(key) is not None
        ]

        num_auto_observe_assets = len(auto_observe_assets)
        has_auto_observe_assets = any(auto_observe_assets)

        if not target_asset_keys and not has_auto_observe_assets:
            self._logger.debug("No assets that require auto-materialize checks")
            yield
            return

        self._logger.info(
            f"Checking {num_target_assets} asset{'' if num_target_assets == 1 else 's'} and"
            f" {num_auto_observe_assets} observable source"
            f" asset{'' if num_auto_observe_assets == 1 else 's'} for auto-materialization"
        )

        raw_cursor = _get_raw_cursor(instance)
        cursor = (
            AssetDaemonCursor.from_serialized(raw_cursor, asset_graph)
            if raw_cursor
            else AssetDaemonCursor.empty()
        )

        tick_retention_settings = instance.get_tick_retention_settings(
            InstigatorType.AUTO_MATERIALIZE
        )

        evaluation_id = cursor.evaluation_id + 1

        tick = instance.create_tick(
            TickData(
                instigator_origin_id=FIXED_AUTO_MATERIALIZATION_ORIGIN_ID,
                instigator_name=FIXED_AUTO_MATERIALIZATION_INSTIGATOR_NAME,
                instigator_type=InstigatorType.AUTO_MATERIALIZE,
                status=TickStatus.STARTED,
                timestamp=evaluation_time.timestamp(),
                selector_id=FIXED_AUTO_MATERIALIZATION_SELECTOR_ID,
                auto_materialize_evaluation_id=evaluation_id,
            )
        )

        with AutoMaterializeLaunchContext(
            tick, instance, self._logger, tick_retention_settings
        ) as tick_context:
            run_requests, new_cursor, evaluations = AssetDaemonContext(
                asset_graph=asset_graph,
                target_asset_keys=target_asset_keys,
                instance=instance,
                cursor=cursor,
                materialize_run_tags={
                    **instance.auto_materialize_run_tags,
                },
                observe_run_tags={AUTO_OBSERVE_TAG: "true"},
                auto_observe=True,
                respect_materialization_data_versions=instance.auto_materialize_respect_materialization_data_versions,
                logger=self._logger,
            ).evaluate()

            self._logger.info(
                f"Tick produced {len(run_requests)} run{'s' if len(run_requests) != 1 else ''} and"
                f" {len(evaluations)} asset evaluation{'s' if len(evaluations) != 1 else ''} for"
                f" evaluation ID {new_cursor.evaluation_id}"
            )

            evaluations_by_asset_key = {
                evaluation.asset_key: evaluation for evaluation in evaluations
            }

            pipeline_and_execution_plan_cache: Dict[
                int, Tuple[ExternalJob, ExternalExecutionPlan]
            ] = {}

            submit_job_inputs: List[Tuple[RunRequest, ExternalJob, ExternalExecutionPlan]] = []

            # First do work that is most likely to fail before doing any writes (to make
            # double-creation of runs less likely)
            for run_request in run_requests:
                yield
                submit_job_inputs.append(
                    get_asset_run_submit_input(
                        run_request._replace(
                            tags={
                                **run_request.tags,
                                AUTO_MATERIALIZE_TAG: "true",
                                ASSET_EVALUATION_ID_TAG: str(new_cursor.evaluation_id),
                            }
                        ),
                        instance,
                        workspace,
                        asset_graph,
                        pipeline_and_execution_plan_cache,
                    )
                )

            tick_context.set_run_requests(run_requests=run_requests)

            # Now submit all runs to the queue
            for submit_job_input in submit_job_inputs:
                yield

                run_request, external_job, external_execution_plan = submit_job_input

                asset_keys = check.not_none(run_request.asset_selection)

                run = submit_asset_run(
                    run_request, instance, workspace, external_job, external_execution_plan
                )

                asset_key_str = ", ".join([asset_key.to_user_string() for asset_key in asset_keys])

                self._logger.info(
                    f"Launched run {run.run_id} for assets {asset_key_str} with tags"
                    f" {run_request.tags}"
                )

                tick_context.add_run_info(run_id=run.run_id)

                # add run id to evaluations
                for asset_key in asset_keys:
                    # asset keys for observation runs don't have evaluations
                    if asset_key in evaluations_by_asset_key:
                        evaluation = evaluations_by_asset_key[asset_key]
                        evaluations_by_asset_key[asset_key] = evaluation._replace(
                            run_ids=evaluation.run_ids | {run.run_id}
                        )

            instance.daemon_cursor_storage.set_cursor_values({CURSOR_KEY: new_cursor.serialize()})
            tick_context.update_state(
                TickStatus.SUCCESS if len(run_requests) > 0 else TickStatus.SKIPPED,
            )

        # We enforce uniqueness per (asset key, evaluation id). Store the evaluations after the cursor,
        # so that if the daemon crashes and doesn't update the cursor we don't try to write duplicates.
        if schedule_storage.supports_auto_materialize_asset_evaluations:
            schedule_storage.add_auto_materialize_asset_evaluations(
                new_cursor.evaluation_id, list(evaluations_by_asset_key.values())
            )
            schedule_storage.purge_asset_evaluations(
                before=pendulum.now("UTC").subtract(days=EVALUATIONS_TTL_DAYS).timestamp(),
            )

        self._logger.info("Finished auto-materialization tick")


def get_asset_run_submit_input(
    run_request: RunRequest,
    instance: DagsterInstance,
    workspace: IWorkspace,
    asset_graph: ExternalAssetGraph,
    pipeline_and_execution_plan_cache: Dict[int, Tuple[ExternalJob, ExternalExecutionPlan]] = {},
) -> Tuple[RunRequest, ExternalJob, ExternalExecutionPlan]:
    check.invariant(
        not run_request.run_config, "Asset materialization run requests have no custom run config"
    )

    asset_keys = check.not_none(run_request.asset_selection)
    check.invariant(len(asset_keys) > 0)

    repo_handle = asset_graph.get_repository_handle(asset_keys[0])

    # Check that all asset keys are from the same repo
    for key in asset_keys[1:]:
        check.invariant(repo_handle == asset_graph.get_repository_handle(key))

    location_name = repo_handle.code_location_origin.location_name
    repository_name = repo_handle.repository_name
    code_location = workspace.get_code_location(location_name)
    job_name = check.not_none(
        asset_graph.get_implicit_job_name_for_assets(
            asset_keys, code_location.get_repository(repository_name)
        )
    )

    job_selector = JobSubsetSelector(
        location_name=location_name,
        repository_name=repository_name,
        job_name=job_name,
        op_selection=None,
        asset_selection=asset_keys,
    )

    selector_id = hash_collection(job_selector)

    if selector_id not in pipeline_and_execution_plan_cache:
        external_job = code_location.get_external_job(job_selector)

        external_execution_plan = code_location.get_external_execution_plan(
            external_job,
            run_config={},
            step_keys_to_execute=None,
            known_state=None,
            instance=instance,
        )
        pipeline_and_execution_plan_cache[selector_id] = (
            external_job,
            external_execution_plan,
        )

    external_job, external_execution_plan = pipeline_and_execution_plan_cache[selector_id]

    return (run_request, external_job, external_execution_plan)


def submit_asset_run(
    run_request: RunRequest,
    instance: DagsterInstance,
    workspace: IWorkspace,
    external_job: ExternalJob,
    external_execution_plan: ExternalExecutionPlan,
):
    asset_keys = check.not_none(run_request.asset_selection)

    execution_plan_snapshot = external_execution_plan.execution_plan_snapshot

    run = instance.create_run(
        job_name=external_job.name,
        run_id=None,
        run_config=None,
        resolved_op_selection=None,
        step_keys_to_execute=None,
        status=DagsterRunStatus.NOT_STARTED,
        op_selection=None,
        root_run_id=None,
        parent_run_id=None,
        tags=run_request.tags,
        job_snapshot=external_job.job_snapshot,
        execution_plan_snapshot=execution_plan_snapshot,
        parent_job_snapshot=external_job.parent_job_snapshot,
        external_job_origin=external_job.get_external_origin(),
        job_code_origin=external_job.get_python_origin(),
        asset_selection=frozenset(asset_keys),
        asset_check_selection=None,
    )
    instance.submit_run(run.run_id, workspace)
    return run



from datetime import datetime
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Union,
    cast,
)

import dagster._check as check
from dagster._annotations import PublicAttr, experimental_param
from dagster._core.definitions.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.events import AssetKey, AssetMaterialization, AssetObservation
from dagster._core.definitions.utils import validate_tags
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.storage.tags import PARTITION_NAME_TAG
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.partition import PartitionsDefinition
    from dagster._core.definitions.run_config import RunConfig
    from dagster._core.definitions.unresolved_asset_job_definition import (
        UnresolvedAssetJobDefinition,
    )


@whitelist_for_serdes(old_storage_names={"JobType"})
class InstigatorType(Enum):
    SCHEDULE = "SCHEDULE"
    SENSOR = "SENSOR"
    AUTO_MATERIALIZE = "AUTO_MATERIALIZE"


@whitelist_for_serdes
class SkipReason(NamedTuple("_SkipReason", [("skip_message", PublicAttr[Optional[str]])])):
    """Represents a skipped evaluation, where no runs are requested. May contain a message to indicate
    why no runs were requested.

    Attributes:
        skip_message (Optional[str]): A message displayed in the Dagster UI for why this evaluation resulted
            in no requested runs.
    """

    def __new__(cls, skip_message: Optional[str] = None):
        return super(SkipReason, cls).__new__(
            cls,
            skip_message=check.opt_str_param(skip_message, "skip_message"),
        )


@whitelist_for_serdes
class AddDynamicPartitionsRequest(
    NamedTuple(
        "_AddDynamicPartitionsRequest",
        [
            ("partitions_def_name", str),
            ("partition_keys", Sequence[str]),
        ],
    )
):
    """A request to add partitions to a dynamic partitions definition, to be evaluated by a sensor or schedule."""

    def __new__(
        cls,
        partitions_def_name: str,
        partition_keys: Sequence[str],
    ):
        return super(AddDynamicPartitionsRequest, cls).__new__(
            cls,
            partitions_def_name=check.str_param(partitions_def_name, "partitions_def_name"),
            partition_keys=check.list_param(partition_keys, "partition_keys", of_type=str),
        )


@whitelist_for_serdes
class DeleteDynamicPartitionsRequest(
    NamedTuple(
        "_AddDynamicPartitionsRequest",
        [
            ("partitions_def_name", str),
            ("partition_keys", Sequence[str]),
        ],
    )
):
    """A request to delete partitions to a dynamic partitions definition, to be evaluated by a sensor or schedule."""

    def __new__(
        cls,
        partitions_def_name: str,
        partition_keys: Sequence[str],
    ):
        return super(DeleteDynamicPartitionsRequest, cls).__new__(
            cls,
            partitions_def_name=check.str_param(partitions_def_name, "partitions_def_name"),
            partition_keys=check.list_param(partition_keys, "partition_keys", of_type=str),
        )


@whitelist_for_serdes
class RunRequest(
    NamedTuple(
        "_RunRequest",
        [
            ("run_key", PublicAttr[Optional[str]]),
            ("run_config", PublicAttr[Mapping[str, Any]]),
            ("tags", PublicAttr[Mapping[str, str]]),
            ("job_name", PublicAttr[Optional[str]]),
            ("asset_selection", PublicAttr[Optional[Sequence[AssetKey]]]),
            ("stale_assets_only", PublicAttr[bool]),
            ("partition_key", PublicAttr[Optional[str]]),
        ],
    )
):
    """Represents all the information required to launch a single run.  Must be returned by a
    SensorDefinition or ScheduleDefinition's evaluation function for a run to be launched.

    Attributes:
        run_key (Optional[str]): A string key to identify this launched run. For sensors, ensures that
            only one run is created per run key across all sensor evaluations.  For schedules,
            ensures that one run is created per tick, across failure recoveries. Passing in a `None`
            value means that a run will always be launched per evaluation.
        run_config (Optional[Mapping[str, Any]]: Configuration for the run. If the job has
            a :py:class:`PartitionedConfig`, this value will override replace the config
            provided by it.
        tags (Optional[Dict[str, Any]]): A dictionary of tags (string key-value pairs) to attach
            to the launched run.
        job_name (Optional[str]): (Experimental) The name of the job this run request will launch.
            Required for sensors that target multiple jobs.
        asset_selection (Optional[Sequence[AssetKey]]): A sequence of AssetKeys that should be
            launched with this run.
        stale_assets_only (bool): Set to true to further narrow the asset
            selection to stale assets. If passed without an asset selection, all stale assets in the
            job will be materialized. If the job does not materialize assets, this flag is ignored.
        partition_key (Optional[str]): The partition key for this run request.
    """

    def __new__(
        cls,
        run_key: Optional[str] = None,
        run_config: Optional[Union["RunConfig", Mapping[str, Any]]] = None,
        tags: Optional[Mapping[str, Any]] = None,
        job_name: Optional[str] = None,
        asset_selection: Optional[Sequence[AssetKey]] = None,
        stale_assets_only: bool = False,
        partition_key: Optional[str] = None,
    ):
        from dagster._core.definitions.run_config import convert_config_input

        return super(RunRequest, cls).__new__(
            cls,
            run_key=check.opt_str_param(run_key, "run_key"),
            run_config=check.opt_mapping_param(
                convert_config_input(run_config), "run_config", key_type=str
            ),
            tags=validate_tags(check.opt_mapping_param(tags, "tags", key_type=str)),
            job_name=check.opt_str_param(job_name, "job_name"),
            asset_selection=check.opt_nullable_sequence_param(
                asset_selection, "asset_selection", of_type=AssetKey
            ),
            stale_assets_only=check.bool_param(stale_assets_only, "stale_assets_only"),
            partition_key=check.opt_str_param(partition_key, "partition_key"),
        )

    def with_replaced_attrs(self, **kwargs: Any) -> "RunRequest":
        fields = self._asdict()
        for k in fields.keys():
            if k in kwargs:
                fields[k] = kwargs[k]
        return RunRequest(**fields)

    def with_resolved_tags_and_config(
        self,
        target_definition: Union["JobDefinition", "UnresolvedAssetJobDefinition"],
        dynamic_partitions_requests: Sequence[
            Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]
        ],
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> "RunRequest":
        from dagster._core.definitions.job_definition import JobDefinition
        from dagster._core.definitions.partition import (
            PartitionedConfig,
            PartitionsDefinition,
        )

        if self.partition_key is None:
            check.failed(
                "Cannot resolve partition for run request without partition key",
            )

        partitions_def = target_definition.partitions_def
        if partitions_def is None:
            check.failed(
                "Cannot resolve partition for run request when target job"
                f" '{target_definition.name}' is unpartitioned.",
            )
        partitions_def = cast(PartitionsDefinition, partitions_def)

        partitioned_config = (
            target_definition.partitioned_config
            if isinstance(target_definition, JobDefinition)
            else PartitionedConfig.from_flexible_config(target_definition.config, partitions_def)
        )
        if partitioned_config is None:
            check.failed(
                "Cannot resolve partition for run request on unpartitioned job",
            )

        _check_valid_partition_key_after_dynamic_partitions_requests(
            self.partition_key,
            partitions_def,
            dynamic_partitions_requests,
            current_time,
            dynamic_partitions_store,
        )

        tags = {
            **(self.tags or {}),
            **partitioned_config.get_tags_for_partition_key(
                self.partition_key,
                job_name=target_definition.name,
            ),
        }

        return self.with_replaced_attrs(
            run_config=(
                self.run_config
                if self.run_config
                else partitioned_config.get_run_config_for_partition_key(self.partition_key)
            ),
            tags=tags,
        )

    def has_resolved_partition(self) -> bool:
        # Backcompat run requests yielded via `run_request_for_partition` already have resolved
        # partitioning
        return self.tags.get(PARTITION_NAME_TAG) is not None if self.partition_key else True


def _check_valid_partition_key_after_dynamic_partitions_requests(
    partition_key: str,
    partitions_def: "PartitionsDefinition",
    dynamic_partitions_requests: Sequence[
        Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]
    ],
    current_time: Optional[datetime] = None,
    dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
):
    from dagster._core.definitions.multi_dimensional_partitions import MultiPartitionsDefinition
    from dagster._core.definitions.partition import (
        DynamicPartitionsDefinition,
    )

    if isinstance(partitions_def, MultiPartitionsDefinition):
        multipartition_key = partitions_def.get_partition_key_from_str(partition_key)

        for dimension in partitions_def.partitions_defs:
            _check_valid_partition_key_after_dynamic_partitions_requests(
                multipartition_key.keys_by_dimension[dimension.name],
                dimension.partitions_def,
                dynamic_partitions_requests,
                current_time,
                dynamic_partitions_store,
            )

    elif isinstance(partitions_def, DynamicPartitionsDefinition) and partitions_def.name:
        if not dynamic_partitions_store:
            check.failed(
                "Cannot resolve partition for run request on dynamic partitions without"
                " dynamic_partitions_store"
            )

        add_partition_keys: Set[str] = set()
        delete_partition_keys: Set[str] = set()
        for req in dynamic_partitions_requests:
            if isinstance(req, AddDynamicPartitionsRequest):
                if req.partitions_def_name == partitions_def.name:
                    add_partition_keys.update(set(req.partition_keys))
            elif isinstance(req, DeleteDynamicPartitionsRequest):
                if req.partitions_def_name == partitions_def.name:
                    delete_partition_keys.update(set(req.partition_keys))

        partition_keys_after_requests_resolved = (
            set(
                dynamic_partitions_store.get_dynamic_partitions(
                    partitions_def_name=partitions_def.name
                )
            )
            | add_partition_keys
        ) - delete_partition_keys

        if partition_key not in partition_keys_after_requests_resolved:
            check.failed(
                f"Dynamic partition key {partition_key} for partitions def"
                f" '{partitions_def.name}' is invalid. After dynamic partitions requests are"
                " applied, it does not exist in the set of valid partition keys."
            )

    else:
        partitions_def.validate_partition_key(
            partition_key,
            dynamic_partitions_store=dynamic_partitions_store,
            current_time=current_time,
        )


@whitelist_for_serdes(
    storage_name="PipelineRunReaction",
    storage_field_names={
        "dagster_run": "pipeline_run",
    },
)
class DagsterRunReaction(
    NamedTuple(
        "_DagsterRunReaction",
        [
            ("dagster_run", Optional[DagsterRun]),
            ("error", Optional[SerializableErrorInfo]),
            ("run_status", Optional[DagsterRunStatus]),
        ],
    )
):
    """Represents a request that reacts to an existing dagster run. If success, it will report logs
    back to the run.

    Attributes:
        dagster_run (Optional[DagsterRun]): The dagster run that originates this reaction.
        error (Optional[SerializableErrorInfo]): user code execution error.
        run_status: (Optional[DagsterRunStatus]): The run status that triggered the reaction.
    """

    def __new__(
        cls,
        dagster_run: Optional[DagsterRun],
        error: Optional[SerializableErrorInfo] = None,
        run_status: Optional[DagsterRunStatus] = None,
    ):
        return super(DagsterRunReaction, cls).__new__(
            cls,
            dagster_run=check.opt_inst_param(dagster_run, "dagster_run", DagsterRun),
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            run_status=check.opt_inst_param(run_status, "run_status", DagsterRunStatus),
        )


@experimental_param(
    param="asset_events", additional_warn_text="Runless asset events are experimental"
)
class SensorResult(
    NamedTuple(
        "_SensorResult",
        [
            ("run_requests", Optional[Sequence[RunRequest]]),
            ("skip_reason", Optional[SkipReason]),
            ("cursor", Optional[str]),
            (
                "dynamic_partitions_requests",
                Optional[
                    Sequence[Union[DeleteDynamicPartitionsRequest, AddDynamicPartitionsRequest]]
                ],
            ),
            (
                "asset_events",
                List[Union[AssetObservation, AssetMaterialization, AssetCheckEvaluation]],
            ),
        ],
    )
):
    """The result of a sensor evaluation.

    Attributes:
        run_requests (Optional[Sequence[RunRequest]]): A list
            of run requests to be executed.
        skip_reason (Optional[Union[str, SkipReason]]): A skip message indicating why sensor
            evaluation was skipped.
        cursor (Optional[str]): The cursor value for this sensor, which will be provided on the
            context for the next sensor evaluation.
        dynamic_partitions_requests (Optional[Sequence[Union[DeleteDynamicPartitionsRequest,
            AddDynamicPartitionsRequest]]]): A list of dynamic partition requests to request dynamic
            partition addition and deletion. Run requests will be evaluated using the state of the
            partitions with these changes applied.
        asset_events (Optional[Sequence[Union[AssetObservation, AssetMaterialization, AssetCheckEvaluation]]]):  (Experimental) A
            list of materializations, observations, and asset check evaluations that the system
            will persist on your behalf at the end of sensor evaluation. These events will be not
            be associated with any particular run, but will be queryable and viewable in the asset catalog.


    """

    def __new__(
        cls,
        run_requests: Optional[Sequence[RunRequest]] = None,
        skip_reason: Optional[Union[str, SkipReason]] = None,
        cursor: Optional[str] = None,
        dynamic_partitions_requests: Optional[
            Sequence[Union[DeleteDynamicPartitionsRequest, AddDynamicPartitionsRequest]]
        ] = None,
        asset_events: Optional[
            Sequence[Union[AssetObservation, AssetMaterialization, AssetCheckEvaluation]]
        ] = None,
    ):
        if skip_reason and len(run_requests if run_requests else []) > 0:
            check.failed(
                "Expected a single skip reason or one or more run requests: received values for "
                "both run_requests and skip_reason"
            )

        skip_reason = check.opt_inst_param(skip_reason, "skip_reason", (SkipReason, str))
        if isinstance(skip_reason, str):
            skip_reason = SkipReason(skip_reason)

        return super(SensorResult, cls).__new__(
            cls,
            run_requests=check.opt_sequence_param(run_requests, "run_requests", RunRequest),
            skip_reason=skip_reason,
            cursor=check.opt_str_param(cursor, "cursor"),
            dynamic_partitions_requests=check.opt_sequence_param(
                dynamic_partitions_requests,
                "dynamic_partitions_requests",
                (AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest),
            ),
            asset_events=list(
                check.opt_sequence_param(
                    asset_events,
                    "asset_check_evaluations",
                    (AssetObservation, AssetMaterialization, AssetCheckEvaluation),
                )
            ),
        )
